(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.

(* Barrier definitions *)

Definition pres (ml : list (assert * assert)) : list assert :=
  map (@fst _ _) ml.

Definition posts (ml : list (assert * assert)) : list assert :=
  map (@snd _ _) ml.

Definition moves_wf (dst : stateid) (bn : bar_num) (cs : stateid) (ml : list (assert * assert)) (limit : nat) : Prop :=
  (forall n P Q, 
    nth_error ml n = Some (P, Q) -> 
      ubiquitous P /\ 
      ubiquitous Q /\ (* Could probably prove this but what a pain *)
      precise P /\
      exists ps, derives P (sepconj (barrier_assert bn ps cs) Tru)) /\
  length ml = limit /\
  exists F,
    (sepconjList (pres ml) = sepconj (barrier_assert bn pfullshare cs) F) /\
    (sepconjList (posts ml) = sepconj (barrier_assert bn pfullshare dst) F).

Record BarMoveList : Type := barMoveList {
  bml_dest : stateid;
  bml_bn : bar_num;
  bml_cs : stateid;
  bml_moves : list (assert * assert);
  bml_limit : nat;
  bml_wf : moves_wf bml_dest bml_bn bml_cs bml_moves bml_limit}.

Definition direction : Type := nat.
Definition move : Type := nat.
Definition directionList : Type := list BarMoveList.

Definition lookup_move' (dl : directionList) (dir move : nat) (pre post : assert) : Prop :=
  match nth_error dl dir with
   | None => False
   | Some (barMoveList _ _ _ moves _ _) =>
     match nth_error moves move with
      | None => False
      | Some (pre', post') => pre = pre' /\ post = post'
     end
  end.

Definition directionswf (bn : bar_num) (cs : stateid) (dl : list BarMoveList) (limit : nat) : Prop :=
  (forall i j, i <> j ->
    forall mi ipre ipost mj jpre jpost,
      lookup_move' dl i mi ipre ipost ->
      lookup_move' dl j mj jpre jpost ->
        (and (sepconj ipre Tru) (sepconj jpre Tru)) = Fls) /\
  (forall i bml, nth_error dl i = Some bml -> 
     bml_cs bml = cs /\
     bml_bn bml = bn /\
     bml_limit bml = limit).

Record BarStateDef : Type := barStateDef {
  bsd_bn : bar_num;
  bsd_cs : stateid;
  bsd_directions : directionList;
  bsd_limit : nat;
  bsd_wf : directionswf bsd_bn bsd_cs bsd_directions bsd_limit}.

Definition bardefwf (bn : bar_num) (lim : nat) (states : list BarStateDef) : Prop :=
  lim > 0 /\
  forall i bsd, 
    nth_error states i = Some bsd -> 
      bsd_limit bsd = lim /\
      bsd_bn bsd = bn /\
      bsd_cs bsd = i.

Record BarDef : Type := barDef {
  bd_bn : bar_num;
  bd_limit : nat;
  bd_states : list BarStateDef;
  bd_wf : bardefwf bd_bn bd_limit bd_states
}.

Definition lookup_move (bsd : BarDef) (cst : stateid) (dir : direction) (move : move) (pre post : assert) : Prop :=
  match nth_error (bd_states bsd) cst with
    | None => False
    | Some bsd => lookup_move' (bsd_directions bsd) dir move pre post
  end.

Lemma lookup_move_fun : forall bd cs rd rm pre post pre' post',
  lookup_move bd cs rd rm pre post ->
  lookup_move bd cs rd rm pre' post' ->
  pre = pre' /\ post = post'.
Proof.
  unfold lookup_move, lookup_move'.
  intros.
  destruct bd.
  icase (nth_error (bd_states (barDef bd_bn0 bd_limit0 bd_states0 bd_wf0))).
  icase (nth_error (bsd_directions b) rd).
  icase b0.
  icase (nth_error bml_moves0 rm).
  icase p.
  destruct H.
  destruct H0.
  split; congruence.
Qed.

Lemma lookup_move_only1DirOK : forall bd cst dir dir' mv mv' pre pre' post post',
  lookup_move bd cst dir mv pre post ->
  lookup_move bd cst dir' mv' pre' post' ->
  (dir = dir') \/ (and (sepconj pre Tru) (sepconj pre' Tru)) = Fls.
Proof.
  intros.
  case (eq_nat_dec dir dir'); intro; auto.
  right.
  destruct bd.
  unfold lookup_move in *.
  remember (nth_error (bd_states (barDef bd_bn0 bd_limit0 bd_states0 bd_wf0)) cst).
  icase e.
  clear -n H H0.
  destruct b.
  simpl in *.
  destruct bsd_wf0.
  spec H1 dir dir' n.
  spec H1 mv  pre post.
  spec H1 mv' pre' post'.
  auto.
Qed.

Lemma lookup_move_facts : forall bd cst dir mv pre post,
  lookup_move bd cst dir mv pre post ->
    ubiquitous pre /\ 
    ubiquitous post /\ 
    precise pre /\
    (exists ps, derives pre (sepconj (barrier_assert (bd_bn bd) ps cst) Tru)).
Proof.
  intros.
  destruct bd.
  unfold lookup_move in H.
  remember (nth_error (bd_states (barDef bd_bn0 bd_limit0 bd_states0 bd_wf0)) cst).
  icase e.
  copy bd_wf0.
  copy bd_wf0.
  destruct bd_wf2.
  spec H1 cst b.
  simpl in Heqe.
  rewrite Heqe in H1.
  detach H1; trivial.
  destruct H1 as [? [? ?]].
  simpl.
  subst.
  clear -H.
  unfold lookup_move' in H.
  destruct b as [? ? ? ? ?].
  simpl in *.
  remember (nth_error bsd_directions0 dir).
  icase e.
  destruct b.
  remember (nth_error bml_moves0 mv).
  icase e.
  destruct p.
  destruct H.
  subst a a0.
  copy bsd_wf0.
  destruct bsd_wf1.
  spec H0 dir (barMoveList bml_dest0 bml_bn0 bml_cs0 bml_moves0 bml_limit0 bml_wf0).
  rewrite Heqe in H0.
  simpl in H0.
  detach H0; trivial.
  destruct H0 as [? [? ?]].
  subst.
  copy bml_wf0.
  destruct bml_wf1 as [? [? [? ?]]].
  spec H0 mv pre post.
  apply H0.
  rewrite Heqe0.
  trivial.
Qed.

Definition lookup_movelist (bsd : BarDef) (cst : stateid) (dir : direction) (preposts : list (assert * assert)) : Prop :=
  match nth_error (bd_states bsd) cst with
    | None => False
    | Some bsd' => match nth_error (bsd_directions bsd') dir with
      | None => False
      | Some (barMoveList _ _ _ moves _ _) => preposts = moves
     end
   end.

Lemma lookup_movelist_fun : forall bsd cst dir pp pp',
  lookup_movelist bsd cst dir pp ->
  lookup_movelist bsd cst dir pp' ->
  pp = pp'.
Proof.
  intros.
  unfold lookup_movelist in *.
  icase (nth_error (bd_states bsd) cst).
  icase (nth_error (bsd_directions b) dir).
  icase b0.
  congruence.
Qed.

Definition lookup_destination (bsd : BarDef) (cst : stateid) (dir : direction) (next : stateid) : Prop :=
  match nth_error (bd_states bsd) cst with
    | None => False
    | Some bsd' => match nth_error (bsd_directions bsd') dir with
      | None => False
      | Some (barMoveList next' _ _ _ _ _) => next = next'
     end
   end.

Lemma lookup_destination_fun : forall bsd cst dir next next',
  lookup_destination bsd cst dir next ->
  lookup_destination bsd cst dir next' ->
  next = next'.
Proof.
  intros.
  unfold lookup_destination in *.
  icase (nth_error (bd_states bsd) cst).
  icase (nth_error (bsd_directions b) dir).
  icase b0.
  congruence.
Qed.

Lemma lookup_movelist_facts: forall bd cst dir dst preposts,
  lookup_movelist bd cst dir preposts ->
  lookup_destination bd cst dir dst ->
  exists F,
    (sepconjList (pres preposts) = sepconj (barrier_assert (bd_bn bd) pfullshare cst) F) /\
    (sepconjList (posts preposts) = sepconj (barrier_assert (bd_bn bd) pfullshare dst) F).
Proof.
  intros.
  destruct bd as [bn lim sts [limx wf]].
  unfold lookup_movelist in H; simpl in H.
  unfold lookup_destination in H0; simpl in H0.
  copy wf.
  remember (nth_error sts cst).
  icase e.
  copy wf.
  spec wf1 cst b.
  rewrite Heqe in wf1.
  detach wf1; trivial.
  destruct wf1 as [? [? ?]].
  subst.
  destruct b as [bn cs dirs lim wf1].
  simpl in *.
  copy wf1.
  destruct wf2.
  remember (nth_error dirs dir).
  icase e.
  spec H2 dir b.
  rewrite Heqe0 in H2.
  detach H2; trivial.
  destruct H2 as [? [? ?]].
  subst.
  destruct b.
  simpl in *.
  copy bml_wf0.
  destruct bml_wf1 as [? [? [? [? ?]]]].
  exists x.
  subst.
  split; trivial.
Qed.

Definition prog : Type :=
  (list BarDef * cmd * list cmd)%type.
