(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.

(* Dynamic concurrent state *)

Definition tid : Type := nat.

Definition slot : Type := (tid * (heap * BarstMap))%type.

Definition slot2state (s : slot) : state :=
  match s with (_, (h, bs)) => (empty_stack, h, bs) end.

Lemma slot2state_zip: forall TL SL,
  length TL = length SL ->
  map slot2state (zip TL SL) = map (fun sx => (empty_stack, fst sx, snd sx)) SL.
Proof.
  induction TL; intros; icase SL.
  simpl in H.
  inversion H.
  spec IHTL SL H1.
  simpl.
  rewrite IHTL.
  destruct p.
  trivial.
Qed.

Definition slot2tid (s : slot) : tid := 
  fst s.

Lemma slot2tid_zip : forall TL SL,
  length TL = length SL ->
  map slot2tid (zip TL SL) = TL.
Proof.
  induction TL; intros; icase SL.
  inversion H.
  simpl.
  rewrite IHTL; auto.
Qed.

Definition wp_getstates (slots : list (option slot)) : list state :=
  map (fun lo => match lo with None => empty_state | Some (_, (hp, bm)) => (empty_stack, hp, bm) end) slots.

Lemma wp_getstates_None : forall s i,
  nth_error s i = None ->
  nth_error (wp_getstates s) i = None.
Proof.
  induction s.
  intros.
  icase i;
  trivial.
  intros.
  destruct i.
  simpl.
  simpl in H.
  inversion H.
  simpl in H.
  simpl.
  auto.
Qed.

Lemma wp_getstates_SomeNone : forall s i,
  nth_error s i = Some None ->
  nth_error (wp_getstates s) i = Some empty_state.
Proof.
  induction s.
  intros.
  icase i.
  intros.
  destruct i.
  simpl in *.
  inversion H.
  trivial.
  simpl.
  apply IHs; trivial.
Qed.

Lemma wp_getstates_SomeSome : forall s i slot,
  nth_error s i = Some (Some slot) ->
  nth_error (wp_getstates s) i = Some (slot2state slot).
Proof.
  induction s.
  intros.
  icase i.
  intros.
  destruct i.
  simpl in *.
  inversion H.
  trivial.
  simpl.
  simpl in H.
  apply IHs.
  trivial.
Qed.

Lemma wp_getstates_eq : forall L L' i,
  nth_error L i = nth_error L' i ->
  nth_error (wp_getstates L) i = nth_error (wp_getstates L') i.
Proof.
  intros until i.
  revert L L'.
  induction i; intros.
  icase L; icase L'.
  inversion H.
  trivial.
  icase L; icase L'; simpl in *.
  spec IHi (@nil (option slot)) L'.
  simpl in IHi.
  rewrite <- IHi; icase i.
  spec IHi L (@nil (option slot)).
  rewrite IHi; icase i.
  auto.
Qed.

Definition wp_wellformed (diro : option direction) (lim : nat) (slots : list (option slot)) (st : state) : Prop :=
  lim = length slots /\
  (diro = None) = (count_optionlist slots = 0) /\
  join_list empty_state (wp_getstates slots) st.

Record Waitpool : Type := waitpool {
  wp_dir : option direction;
  wp_slots : list (option slot);
  wp_limit : nat;
  wp_st : state;
  wp_wf : wp_wellformed wp_dir wp_limit wp_slots wp_st}.

Lemma empty_wp_ok: forall lim, lim > 0 ->
  wp_wellformed None lim (n_None slot lim) empty_state.
Proof.
  unfold wp_wellformed.
  intros.
  split.
  clear H.
  induction lim; auto.
  simpl.
  rewrite <- IHlim.
  trivial.
  split.
  apply prop_ext.
  split; intro; auto.
  clear H.
  induction lim; auto.
  clear H.
  induction lim; auto.
  constructor.
  simpl.
  econstructor.
  apply join_emptystate; trivial.
  trivial.
Qed.  

Definition waiting (wp : Waitpool) : nat :=
  count_optionlist (wp_slots wp).

Definition dbs_wellformed (bn : bar_num) (cs : stateid) (wp : Waitpool) (bd : BarDef) : Prop :=
  match wp with waitpool diro slots lim _ _ =>
     lim = bd_limit bd /\
     bn = bd_bn bd /\
     match diro with 
      | None => forall i, i < lim -> nth_error slots i = Some None
      | Some dir => 
        forall i, match nth_error slots i with
          | None => True
          | Some None => True
          | Some (Some slt) => 
             exists pre, exists post, 
               lookup_move bd cs dir i pre post /\
               pre (slot2state slt)
     end
     end
  end.

Record DyBarStatus : Type := dyBarStatus {
  dbs_bn : bar_num;
  dbs_cs : stateid;
  dbs_wp : Waitpool;
  dbs_bd : BarDef;
  dbs_wf : dbs_wellformed dbs_bn dbs_cs dbs_wp dbs_bd}.

Definition dbs_inc_waitpool 
        (dbs : DyBarStatus) (dir : direction) (mn : move) (nv : slot) 
	(dbs' : DyBarStatus) : Prop :=
  match (dbs, dbs') with 
    (dyBarStatus bn cs (waitpool diro slots lim st _) bd _, 
     dyBarStatus bn' cs' (waitpool diro' slots' lim' st' _) bd' _) =>
       bn = bn' /\
       cs = cs' /\
       (diro = None \/ diro = Some dir) /\
       diro' = Some dir /\
       replace slots mn (Some nv) = Some slots' /\
       lim = lim' /\
       join (slot2state nv) st st' /\
       bd = bd'
  end.

Lemma can_dbs_inc_wp: forall diro dir bd cs mn pre post st slots slot bn lim wfwp wfdbs,
  (diro = None \/ diro = Some dir) ->
  lookup_move bd cs dir mn pre post ->
  joins (slot2state slot) st ->
  pre (slot2state slot) ->
  exists dbs',
  dbs_inc_waitpool (dyBarStatus bn cs (waitpool diro slots lim st wfwp) bd wfdbs) 
                                 dir mn slot 
                                 dbs'.
Proof.
  intros.
  copy wfdbs.
  destruct wfdbs as [? [? ?]].
  copy wfwp.
  copy wfwp.
  destruct wfwp1 as [? [? ?]].
  assert (exists osv, nth_error slots mn = Some osv).
    apply can_nth_error.
    rewrite <- H6.
    rewrite H3.
    destruct bd as [bn' lim' sts wfbd].
    simpl.
    simpl in H3, H4.
    subst bn' lim'.
    copy wfbd.
    unfold lookup_move in H0.
    simpl in H0.
    remember (nth_error sts cs).
    icase e.
    symmetry in Heqe.
    destruct wfbd0.
    spec H4 cs b Heqe.
    destruct H4 as [? [? ?]].
    destruct b as [bn' cs' dl lim' bsdwf].
    simpl in H4, H9, H10.
    subst lim' bn' cs'.
    copy bsdwf.
    copy bsdwf.
    destruct bsdwf1.
    unfold lookup_move' in H0.
    simpl in H0.
    remember (nth_error dl dir).
    icase e.
    spec H9 dir b.
    rewrite <- Heqe0 in H9.
    detach H9; trivial.
    destruct H9 as [? [? ?]].
    rewrite <- H11.
    destruct b.
    simpl.
    remember (nth_error bml_moves mn).
    icase e.
    copy bml_wf.
    destruct bml_wf0 as [? [? [? [? ?]]]].
    rewrite <- H13.
    apply nth_error_bound with p.
    symmetry.
    trivial.
  destruct H9 as [osv ?].
  destruct (can_replace _ _ _ _ H9 (Some slot0)) as [slots' ?].
  destruct H1 as [st' ?].
  assert (wp_wellformed (Some dir) lim slots' st').
    unfold wp_wellformed.
    split.
    generalize (length_replace _ _ _ _ _ H10); congruence.
    split.
    apply prop_ext.
    split; intro; disc.
    elimtype False.
    clear -H11 H10.
    revert mn slots' slot0 H10 H11.
    induction slots; simpl; intros.
    icase mn.
    icase mn.
    inversion H10.
    rewrite <- H0 in H11.
    simpl in H11.
    omega.
    remember (replace slots mn (Some slot0)).
    icase o.
    spec IHslots mn l slot0.
    rewrite Heqo in IHslots.
    apply IHslots; auto.
    inversion H10.
    rewrite <- H0 in H11.
    simpl in H11.
    destruct a.
    omegac.
    trivial.
(* this next bit is nasty *)
    assert (exists oh, nth_error (wp_getstates slots) mn = Some oh).
      apply can_nth_error.
      unfold wp_getstates.
      rewrite map_length.
      apply nth_error_bound with osv.
      trivial.
    destruct H11 as [ost ?].
    assert (st_stk ost = empty_stack).
      unfold wp_getstates in H11.
      rewrite nth_error_map in H11.
      change (prod tid (prod heap BarstMap)) with slot in H11.
      rewrite H9 in H11.
      simpl in H11.
      unfold empty_state in H11.
      destruct ost.
      destruct p.
      destruct osv.
      destruct s0.
      destruct p.
      inversion H11.
      trivial.
      inversion H11.
      trivial.
    destruct (@join_list_split_nth _ _ _ _ _ _ _ _ _ H11 H8 (join_emptystate ost H12)) as [? [sl [? [? ?]]]].
    generalize (empty_state_identity _ _ H14); intro.
    subst x.
    clear H14.
    assert (ost = empty_state).
      destruct H; subst.
      (* leg 1 *)
      spec H5 mn.
      rewrite H6 in H5.
      spec H5 (nth_error_bound _ _ _ _ H9).
      rewrite (wp_getstates_SomeNone _ _ H5) in H11.
      inversion H11.
      trivial.
      (* leg 2 *)
      spec H5 mn.
      rewrite H9 in H5.
      destruct osv.
      (* leg 2a *)
      destruct (lookup_move_facts _ _ _ _ _ _ H0) as [? [? [? ?]]].
      destruct H5 as [pre' [post' [? ? ]]].
      destruct (lookup_move_fun _ _ _ _ _ _ _ _ H5 H0).
      subst pre' post'.
      assert (join empty_state (slot2state s) (slot2state s)).
        apply join_emptystate.
        destruct s.
        destruct p.
        trivial.
      generalize (wp_getstates_SomeSome _ _ _ H9); intro.
      destruct (@join_list_split_nth _ _ _ _ _ _ _ _ _ H18 H8 H17) as [? [sl' [? [? ?]]]].
      generalize (empty_state_identity _ _ H20); intro.
      subst x.
      destruct (join_list_assoc2 H21 H1) as [? [? ?]].
      spec H4 x (slot2state slot0) (slot2state s).
      detach H4.
      detach H4.
      spec H4 H2 H16.
      destruct slot0 as [? [? ?]].
      destruct s as [? [? ?]].
      inversion H4.
      subst h0 b0.
      generalize (join_self _ _ H22); intro.
      subst x.
      unfold slot2state in H22.
      assert (unit_for (empty_stack, h, b) (empty_stack, h, b)) by trivial.
      rewrite <- identity_unit_equiv in H24.
      apply identity_emptystate in H24; trivial.
      unfold empty_state in H24.
      inversion H24.
      subst h b.
      assert (st = st').
        unfold slot2state in H21, H23.
        apply join_list_eq with empty_state sl'; trivial.
      subst st'.
      rewrite H11 in H18.
      inversion H18.
      trivial.
      exists (slot2state slot0).
      apply join_com.
      trivial.
      exists (slot2state s).
      trivial.
      (* leg 2b, like leg 1 *)
      rewrite (wp_getstates_SomeNone _ _ H9) in H11.
      inversion H11.
      trivial.
    subst ost.
    generalize (replace_nth _ _ _ _ _ H13); intro.
    apply join_com in H1.
    destruct (@join_list_join_nth _ _ _ _ _ _ _ _ _ H14 H15 H1) as [? [sl' [? [? ?]]]].
    generalize (empty_state_identity _ _ H16); intro.
    subst x.
    replace (wp_getstates slots') with sl'; trivial.
    apply nth_error_eq.
    intro.
    case (eq_nat_dec mn i); intro.
    subst i.
    generalize (replace_nth _ _ _ _ _ H10); intro.
    generalize (replace_nth _ _ _ _ _ H17); intro.
    rewrite (wp_getstates_SomeSome _ _ slot0); trivial.
    generalize (replace_nth2 _ _ _ _ _ _ H10 n); intro.
    generalize (replace_nth2 _ _ _ _ _ _ H17 n); intro.
    generalize (replace_nth2 _ _ _ _ _ _ H13 n); intro.
    rewrite H20.
    rewrite H21.
    apply wp_getstates_eq.
    symmetry.
    trivial.
  assert (dbs_wellformed bn cs (waitpool (Some dir) slots' lim st' H11) bd).
    unfold dbs_wellformed.
    split; trivial.
    split; trivial.
    intro.
    case (eq_nat_dec mn i); intro.
    subst.
    rewrite (replace_nth _ _ _ _ _ H10).
    exists pre.
    exists post.
    split; trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H10 n).
    destruct H; subst.
    remember (nth_error slots i).
    destruct e; trivial.
    symmetry in Heqe.
    generalize (nth_error_bound _ _ _ _ Heqe); intro.
    rewrite H6 in H5.
    spec H5 i H.
    rewrite Heqe in H5.
    inversion H5; trivial.
    apply H5.
  exists (dyBarStatus bn cs (waitpool (Some dir) slots' lim st' H11) bd H12).
  repeat (split; trivial).
Qed.

Record DbsTransOut : Type := dbsTransOut {
  dbsto_src : stateid;
  dbsto_dir : direction;
  dbsto_dst : stateid;
  dbsto_st : state;
  dbsto_slots : list slot;
  dbsto_wf : join_list empty_state (map slot2state dbsto_slots) dbsto_st}.

Definition dbs_transition 
        (dbs : DyBarStatus)
	(dbs' : DyBarStatus) (out : DbsTransOut) : Prop :=
  match (dbs, dbs') with 
    (dyBarStatus bn cs (waitpool diro slots lim st _) bd _, 
     dyBarStatus bn' cs' (waitpool diro' slots' lim' st' _) bd' _) =>
     bn = bn' /\
     exists dir,
       diro = Some dir /\
       diro' = None /\
       lookup_destination bd cs dir cs' /\
       slots' = n_None slot lim /\
       lim = lim' /\
       st' = empty_state /\
       bd = bd' /\
       (exists F, (sepconj (barrier_assert bn pfullshare cs) F) st) /\
       match out with dbsTransOut src dirx dst st'' newslots _ =>
         dirx = dir /\
         src = cs /\
         dst = cs' /\
         st'' = st_upd_bm_bn st bn (barst_builder dst pfullshare) /\
         (forall pl, lookup_movelist bd cs dir pl ->
             assert_list_state_list (posts pl) 
                                                (map slot2state newslots)) /\
         dualolist (map (option_map slot2tid) slots) 
                        (map slot2tid newslots)
       end
  end.

Definition dbs_ready (dbs : DyBarStatus) : Prop :=
  fullolist (wp_slots (dbs_wp dbs)).

Lemma can_dbs_transition: forall dbs,
  dbs_ready dbs ->
  exists dbs', exists out,
    dbs_transition dbs dbs' out.
Proof.
  intros [bn cs [diro slots lim st wfwp] bd wfdbs] ?.
  unfold dbs_ready in H.
  simpl in H.
  copy wfdbs.
  destruct wfdbs as [? [? ?]].
  destruct bd as [bn' lim' sts wfbd].
  simpl in H0, H1.
  subst lim' bn'.
  destruct diro.
  rename d into dir.
  copy wfwp.
  copy wfwp.
  destruct wfwp1 as [? [? ?]].
  copy H2.
  spec H4 0.
  remember (nth_error slots 0).
  destruct e.
  2: destruct slots; simpl in *; disc; copy wfbd; copy wfbd; destruct wfbd1; omegac.
  destruct o.
  2: (destruct slots; inversion Heqe; subst o; contradiction (None_fullolist _ slots)).
  destruct H4 as [pre [post [? ?]]].
  clear H5.
  unfold lookup_move, lookup_move' in H4.
  simpl in H4.
  remember (nth_error sts cs).
  icase e.
  destruct b as [bn' cs' dirs lim' wfbs].
  simpl in H4.
  copy wfbd.
  copy wfbd.
  destruct wfbd1.
  spec H6 cs (barStateDef bn' cs' dirs lim' wfbs).
  symmetry in Heqe0.
  spec H6 Heqe0.
  simpl in H6.
  destruct H6 as [? [? ?]].
  subst lim' bn' cs'.
  copy wfbs.
  destruct wfbs0.
  remember (nth_error dirs dir).
  icase e.
  symmetry in Heqe1.
  spec H7 dir b Heqe1.
  destruct b as [dst bn' cs' mvs lim' wfbml].
  simpl in H7.
  destruct H7 as [? [? ?]].
  subst lim' cs' bn'.

  copy wfbml.
  destruct wfbml0 as [? [? [F [? ?]]]].
  destruct (fullo_dualo _ _ H) as [out0 ?].
  generalize (dualo_length _ _ _ H11); intro dlen.
  assert (assert_list_state_list (pres mvs) (map slot2state out0)).
    unfold lookup_move, lookup_move' in H2.
    simpl in H2.
    rewrite Heqe0 in H2.
    simpl in H2.
    rewrite Heqe1 in H2.
    clear -H2 H11 H8 H0.
    revert slots H0 mvs H8 out0 H11 H2.
    unfold dualolist.
    (* Yikes! *)
    induction lim; intros.
    destruct slots; simpl in *; disc.
    destruct out0; simpl in *; disc.
    destruct mvs; simpl in *; disc.
    trivial.
    destruct slots; simpl in *; disc.
    destruct out0; simpl in *; disc.
    destruct mvs; simpl in *; disc.
    copy H2.
    spec H2 0; simpl in H2.
    destruct p.
    icase o.
    destruct H2 as [pre [post [[? ?] ?]]].
    subst a a0.
    simpl.
    inversion H11.
    subst s0.
    split; trivial.
    inversion H0.
    clear H0.
    inversion H8.
    clear H8.
    spec IHlim slots H2 mvs H0.
    spec IHlim out0 H4.
    apply IHlim.
    clear IHlim.
    intro.
    spec H1 (S i).
    simpl in H1.
    trivial.
  assert ((sepconjList (pres mvs)) st).
    apply assertlist_statelist_sepconjlist with 
      (map slot2state out0) empty_state; auto.
    apply empty_state_identity.
    clear -H11 H3.
    remember empty_state.
    clear Heqs.
    revert out0 H11 s st H3.
    induction slots; intros.
    icase out0.
    icase out0.
    icase a.
    inversion H11.
    inversion H3; subst.
    destruct s0.
    destruct p.
    econstructor.
    apply H5.
    spec IHslots out0.
    detach IHslots.
    2: compute; trivial.
    spec IHslots w1 st H7.
    trivial.
  (* Do the switch! *)
  rewrite H9 in H13.
  destruct st as [[? ?] ?].
  set (b' := upd_bm b bn (barst_builder dst pfullshare)).
  rename H13 into H13'.
  assert ((sepconj (barrier_assert bn pfullshare dst) F) (s0, h, b')).
    clear -H13'.
    inv_sc H13'.
    subst s1 s0.
    destruct H3 as [? [? ?]].
    unfold st_bsm, st_hp, fst, snd in H, H3.
    subst h0.
    generalize (empty_heap_identity _ _ H1); intro.
    subst h1.
    clear H1.
    exists (s, empty_heap, upd_bm b0 bn (barst_builder dst pfullshare)).
    exists (s, h, b1).
    split.
    split.
    split.
    split; trivial.
    apply join_emptyheap.
    intro.
    unfold b', upd_bm, snd.
    spec H2 x.
    case (eq_nat_dec x bn); intro.
    subst x.
    icase (b1 bn).
    rewrite H in H2.
    inversion H2.
    subst.
    destruct p.
    destruct z.
    unfold pjoin in H6.
    destruct x.
    destruct x0.
    destruct H6.
    unfold snd in H3.
    change Share.top with (projT1 pfullshare) in H3.
    generalize (nidentity_pair2 n); intro.
    change t with (projT1 (existT (fun t => ~identity t) _ H6)) in H3.
    pfullshare_join.
    constructor.
    trivial.
    split; trivial.
    unfold barrier_assert, st_bsm, snd, upd_bm.
    split.
    rewrite if_true; auto.
    split.
    intros.
    rewrite if_false; auto.
    trivial.
  rewrite <- H10 in H13.
  assert (s0 = empty_stack).
    clear -H3.
    unfold empty_state in H3.
    remember empty_heap.
    remember empty_bm.
    clear Heqh0 Heqb0.
    revert h0 b0 H3.
    induction slots; intros.
    inversion H3; auto.
    inversion H3; subst.
    destruct a.
    destruct s.
    destruct p.
    destruct w1 as [[? ?] ?].
    destruct H2.
    destruct H.
    destruct H.
    unfold fst, snd in *.
    subst s.
    spec IHslots h2 b2 H5.
    trivial.
    apply join_com in H2.
    generalize (empty_state_identity _ _ H2); intro.
    destruct w1 as [[? ?] ?].
    inversion H.
    subst.
    spec IHslots h1 b1.
    auto.
  subst s0.
  destruct (sepconjList_joinlist _ _ _ _ H13) as [sl2 [? ?]].
  assert (length sl2 = length mvs).
    clear -H15.
    replace (length mvs) with (length (posts mvs)).
    2: unfold posts; rewrite map_length; trivial.
    revert sl2 H15.
    induction mvs; intros.
    simpl in H15.
    icase sl2.
    icase sl2.
    simpl in H15.
    destruct H15.
    simpl.
    rewrite IHmvs; trivial.
  rewrite H8 in H16.
  assert (dbs_wellformed bn dst (waitpool None (n_None slot lim) lim empty_state (empty_wp_ok lim H5)) (barDef bn lim sts wfbd0)).
    split; trivial.
    split; trivial.
    clear.
    induction lim; intros.
    inversion H.
    simpl.
    destruct i; auto.
    assert (i < lim) by omega.
    spec IHlim i H0.
    apply IHlim.
  exists (dyBarStatus bn dst (waitpool None (n_None slot lim) lim empty_state (empty_wp_ok lim H5)) (barDef bn lim sts wfbd0) H17).
  assert (map (fun x : stack * heap * BarstMap => (empty_stack, snd (fst x), snd x)) sl2 = sl2).
    clear -H14.
    apply nth_error_eq.
    intro.
    rewrite nth_error_map.
    remember (nth_error sl2 i).
    icase e.
    destruct p as [[? ?] ?].
    simpl.
    replace s with empty_stack; trivial.
    remember empty_heap.
    remember empty_bm.
    clear Heqh1 Heqb1.
    revert h1 b1 H14 i Heqe.
    induction sl2; intros.
    icase i.
    destruct i.
    inversion Heqe.
    destruct a as [[? ?] ?].
    inversion H0.
    subst.
    inversion H14.
    destruct H3 as [[[? ?] _] _].
    apply H3.
    simpl in Heqe.
    inversion H14; subst.
    destruct w1 as [[? ?] ?].
    destruct H2 as [[[? ?] _] _].
    simpl in H, H0.
    rewrite <- H in H0.
    subst s0.
    spec IHsl2 h2 b2 H4.
    apply IHsl2 with i; auto.
(*  generalize (joinlistState_joinlistHeap _ _ _ _ _ H13); intro. *)
  set (olist := zip (map (@fst _ _) out0) (map (fun st => (snd (fst st), snd st)) sl2)).
  assert (join_list empty_state (map slot2state olist) (empty_stack, h,b')).
    unfold olist.
    rewrite slot2state_zip.
    rewrite map_comp'.
    simpl.
    rewrite H18.
    trivial.
    do 2 rewrite map_length.
    rewrite H16.
    change (prod tid (prod heap BarstMap)) with slot.
    rewrite <- dlen.
    auto.
  exists (dbsTransOut cs dir dst (empty_stack, h, b') olist H19).
  split; trivial.
  exists dir.
  unfold lookup_destination; simpl.
  rewrite Heqe0.
  simpl.
  rewrite Heqe1.
  do 8 (split; trivial).
  exists F.
  trivial.
  do 5 (split; trivial).
  intros.
  unfold lookup_movelist in H20.
  simpl in H20.
  rewrite Heqe0 in H20.
  simpl in H20.
  rewrite Heqe1 in H20.
  subst pl.
  unfold olist.
  rewrite slot2state_zip.
  rewrite map_comp'.
  simpl.
  rewrite H18.
  trivial.
  do 2 (rewrite map_length).
  rewrite H16.
  change (prod tid (prod heap BarstMap)) with slot.
  rewrite <- dlen.
  auto.
  unfold olist.
  rewrite slot2tid_zip.
  apply dualo_map.
  trivial.
  do 2 (rewrite map_length).
  rewrite H16.
  change (prod tid (prod heap BarstMap)) with slot.
  rewrite <- dlen.
  auto.

  copy wfwp.
  copy wfwp.
  destruct wfwp1 as [? [? ?]].
  unfold fullolist in H.
  rewrite H in H1.
  rewrite <- H0 in H1.
  assert (lim = 0) by (rewrite <- H1; trivial).
  copy wfbd.
  copy wfbd.
  destruct wfbd1.
  omegac.
Qed.

Definition dbs_getsts : list DyBarStatus -> list state :=
  map (wp_st oo dbs_wp).

Definition bp_wellformed (bars : list DyBarStatus) (st : state) : Prop :=
  join_list empty_state (dbs_getsts bars) st /\
  forall n dbs, nth_error bars n = Some dbs ->
    dbs_bn dbs = n.

Record Barpool : Type := barpool {
  bp_bars : list DyBarStatus;
  bp_st : state;
  bp_wf : bp_wellformed bp_bars bp_st}.

Definition bp_inc_waitpool (bp : Barpool) (bn : bar_num) (dir : direction)
                                              (mn : move) (nv : slot) (bp' : Barpool) : Prop :=
  match (bp, bp') with
    (barpool bars st' _, barpool bars' st'' _) =>
    join (slot2state nv) st' st'' /\
      match (nth_error bars bn, nth_error bars' bn) with
        | (None, _) => False
        | (_, None) => False
        | (Some dbs, Some dbs') => 
          dbs_inc_waitpool dbs dir mn nv dbs' /\
          replace bars bn dbs' = Some bars'
      end
  end.

Lemma can_bp_inc_wp: forall bp bn wp dir bd cs mn pre post hp bs tid wfdbs,
  nth_error (bp_bars bp) bn = 
    Some (dyBarStatus bn cs wp bd wfdbs) ->
  (wp_dir wp = None \/ wp_dir wp = Some dir) ->
  lookup_move bd cs dir mn pre post ->
  joins (empty_stack, hp, bs) (bp_st bp) ->
  pre (empty_stack, hp, bs) ->
  exists bp',
    bp_inc_waitpool bp bn dir mn (tid, (hp, bs)) bp'.
Proof.
  intros.
  destruct bp as [bars bst wfbp].
  destruct wp as [diro slots lim wpst wpwf].
  simpl in H0, H.
  unfold bp_st in H2.
  assert (nth_error (dbs_getsts bars) bn = Some wpst).
    unfold dbs_getsts.
    clear -H.
    (* ugly... *)
    remember bn.
    pattern n at 1 in H.
    rewrite Heqn in H.
    rewrite Heqn.
    clear Heqn.
    (* end ugly *)
    revert bn H.
    induction bars; intros.
    icase bn.
    destruct bn.
    inversion H.
    trivial.
    simpl.
    simpl in H.
    spec IHbars bn H.
    trivial.
  destruct H2 as [st'' H2].
  apply join_com in H2.
  copy wfbp.
  destruct wfbp.
  destruct (@join_list_join_nth _ _ _ _ _ _ _ _ _ H4 H5 H2) as [nwphp [barshps'' [? [? ?]]]].
  apply join_com in H7.
  assert (joins (slot2state (tid0, (hp, bs))) wpst).
    exists nwphp; trivial.
  destruct (can_dbs_inc_wp _ _ _ _ _ _ _ _ _ _ _ _ wpwf wfdbs H0 H1 H10 H3) as [dbs' ?].
  copy H11.
  destruct dbs' as [bn' cs' [diro' slots' lim' st''2 wfwp'] bd' wfdbs'].
  destruct H11 as  [? [? [? [? [? [? [? ?]]]]]]].  
  subst bn' cs' diro' lim' bd'.
  unfold slot2state in H18.
  generalize (join_eq _ _ _ _ H7 H18); intro.
  subst st''2.
  destruct (can_replace _ _ _ _ H 
                 (dyBarStatus bn cs (waitpool (Some dir) slots' lim nwphp wfwp') bd wfdbs')) as 
                 [bars' ?].
  assert (join_list empty_state (dbs_getsts bars') st'').
    replace (dbs_getsts bars') with barshps''; trivial.
    apply nth_error_eq.
    intro.
    case (eq_nat_dec bn i); intro.
    subst i.
    rewrite (replace_nth _ _ _ _ _ H8).
    unfold dbs_getsts.
    rewrite nth_error_map.
    rewrite (replace_nth _ _ _ _ _ H11).
    trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H8 n).
    unfold dbs_getsts.
    do 2 rewrite nth_error_map.
    rewrite (replace_nth2 _ _ _ _ _ _ H11 n).
    trivial.
  assert (bp_wellformed bars' st'').
    split; trivial.
    intros.
    spec H6 n dbs.
    case (eq_nat_dec bn n); intro.
    subst n.
    rewrite (replace_nth _ _ _ _ _ H11) in H15.
    inversion H15.
    trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H11 n0) in H15.
    rewrite H15 in H6.
    auto.
  exists (barpool bars' st'' H15).
  split.
  apply join_com.
  trivial.
  rewrite H.
  rewrite (replace_nth _ _ _ _ _ H11).
  split; trivial.
Qed.

Definition bp_transition (bp : Barpool) (bn : bar_num)
                                         (bp' : Barpool) (out : DbsTransOut) : Prop :=
  match (bp, bp', out) with
    (barpool bars st'' _, barpool bars' st' _, dbsTransOut src dirx dst st slots _) =>
    (st_bsm st'' bn = barst_builder src pfullshare) /\
    join st st' (st_upd_bm_bn st'' bn (barst_builder dst pfullshare)) /\
      match (nth_error bars bn, nth_error bars' bn) with
        | (None, _) => False
        | (_, None) => False
        | (Some dbs, Some dbs') => 
          dbs_transition dbs dbs' out /\
          replace bars bn dbs' = Some bars'
      end
  end.

Definition bp_ready (bp : Barpool) (bn : bar_num) :=
  exists dbs,
    nth_error (bp_bars bp) bn = Some dbs /\
    dbs_ready dbs.

Lemma can_bp_transition: forall bp bn,
  bp_ready bp bn ->
  exists bp', exists out, 
    bp_transition bp bn bp' out.
Proof.
  intros.
  destruct H as [dbs [H H0]].
  destruct bp as [bars bhp wfbp].
  destruct (can_dbs_transition dbs H0) as [dbs' [out ?]].
  unfold bp_bars in H.
  destruct (can_replace _ _ _ _ H dbs') as [bars' ?].
  destruct dbs as [bn' cs wp bd wfdbs].
  destruct dbs' as [bn'' cs' wp' bd' wfdbs'].
  copy H1.
  destruct wp as [diro slots lim st wfwp].
  destruct wp' as [diro' slots' lim' st' wfwp'].
  destruct H3 as [? [? [? [? [? [? [? [? [? ?]]]]]]]]].
  subst bn'' diro diro' slots' lim' st' bd'.
  copy wfbp.
  destruct wfbp.
  spec H4 bn (dyBarStatus bn' cs (waitpool (Some x) slots lim st wfwp) bd wfdbs) H.
  simpl in H4.
  subst bn'.
  assert (nth_error (dbs_getsts bars) bn = Some st).
    unfold dbs_getsts.
    clear -H.
    remember bn.
    rewrite Heqb.
    pattern b at 1 in H.
    rewrite Heqb in H.
    clear Heqb.
    revert bn H.
    induction bars; intros.
    icase bn.
    destruct bn.
    inversion H.
    trivial.
    simpl.
    simpl in H.
    spec IHbars bn H.
    trivial.
  assert (join empty_state st st).
    apply join_emptystate.
    clear -H4 H3.
    unfold empty_state in H3.
    remember empty_heap.
    remember empty_bm.
    clear Heqh Heqb.
    revert bn H4 h b H3.
    induction bars; intros; icase bn.
    inversion H3.
    inversion H4.
    subst.
    destruct a.
    unfold compose in *.
    simpl in *.
    destruct dbs_wp0.
    simpl in *.
    destruct wp_st0 as [[? ?] ?].
    unfold st_stk.
    simpl.
    destruct H2 as [[[? ?] _] _].
    auto.
    simpl in H4.
    inversion H3; subst.
    destruct w1 as [[? ?] ?].
    spec IHbars bn H4 h0 b0.
    apply IHbars.
    replace empty_stack with s; trivial.
    destruct H2 as [[[? ?] _] _].
    rewrite H0 in H.
    auto.
  destruct (@join_list_split_nth _ _ _ _ _ _ _ _ _ H4 H3 H5) as [st' [sx [? [? ?]]]].
  generalize (empty_state_identity _ _ H8); intro.
  subst st'.
  clear H8.

  assert (dbs_getsts bars' = sx).
    clear -H2 H7.
    apply nth_error_eq; intro.
    unfold dbs_getsts in *.
    rewrite nth_error_map.
    case (eq_nat_dec bn i); intro.
    subst i.
    generalize (replace_nth _ _ _ _ _ H7); intro.
    generalize (replace_nth _ _ _ _ _ H2); intro.
    rewrite H0, H.
    compute.
    trivial.
    generalize (replace_nth2 _ _ _ _ _ _ H7 n); intro.
    generalize (replace_nth2 _ _ _ _ _ _ H2 n); intro.
    rewrite H, H0.
    rewrite nth_error_map.
    trivial.
  subst sx.
  unfold wp_st in *.
  destruct (@join_list_assoc1 _ _ _ _ _ _ _ (join_com _ _ _ H5) H9) as [bpst' [? ?]].
  assert (bp_wellformed bars' bpst').
    split; trivial.
    copy wfbp0.
    destruct wfbp1.
    intros.
    spec H13 n dbs.
    case (eq_nat_dec bn n); intro.
    subst n.
    rewrite (replace_nth _ _ _ _ _ H2) in H14.
    inversion H14.
    trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H2 n0) in H14.
    auto.
  exists (barpool _ _ H12).
  exists out.
  unfold bp_transition.
  destruct out.
  destruct H11 as [[F ?] [? [? [? [? ?]]]]].
  subst dbsto_dir0 dbsto_st0 dbsto_src0 dbsto_dst0.
  split; trivial.
  clear -H10 H11.
  destruct H10 as [? ?].
  spec H0 bn.
  destruct st as [[? ?] ?].
  inv_sc H11.
  destruct bpst' as [[? ?] ?].
  destruct bhp as [[? ?] ?].
  destruct H5 as [? [? ?]].
  unfold st_bsm, snd in *.
  spec H4 bn.
  destruct (join_assoc _ _ _ _ _ H4 H0) as [? [? ?]].
  rewrite H5 in H10.
  inversion H10.
  trivial.
  destruct y.
  destruct z.
  simpl in H14.
  destruct H14.
  unfold snd in H15.
  destruct x1.
  destruct x2.
  change Share.top with (projT1 pfullshare) in H15.
  generalize (nidentity_pair2 n); intro.
  change t with (projT1 (existT (fun x => ~identity x)  _ H16)) in H15.
  pfullshare_join.
  destruct H10 as [? ?].
  split.
  split; trivial.
  unfold st_upd_bm_bn, upd_bm, snd.
  destruct st as [[? ?] ?].
  intro.
  spec H13 x0.
  case (eq_nat_dec x0 bn); intro.
  subst x0.
  destruct bpst'.
  unfold fst, snd in *.
  remember (b0 bn).
  destruct b1.
  elimtype False.
  destruct bhp.
  clear -H11 H13.
  inv_sc H11.
  subst s0 s1.
  spec H2 bn.
  destruct H3 as [? [? ?]].
  unfold st_bsm, fst, snd in H.
  apply join_com in H2.
  destruct (join_assoc _ _ _ _ _ H2 H13) as [? [? ?]].
  rewrite H in H5.
  clear -H5.
  inversion H5.
  subst.
  destruct p0.
  destruct z.
  simpl in H1.
  destruct H1.
  simpl in H0.
  generalize (nidentity_pair2 n); intro.
  change (snd x) with (projT1 (existT (fun x => ~identity x)  _ H1)) in H0.
  change Share.top with (projT1 pfullshare) in H0.
  pfullshare_join.
  constructor.
  apply H13.
  rewrite H.
  rewrite (replace_nth _ _ _ _ _ H2).
  split; trivial.
Qed.

Definition lookup_move_bp (bp : Barpool) (bn : bar_num) (dir : direction) (mv : move) (pre post : assert) : Prop :=
  match nth_error (bp_bars bp) bn with
    | None => False
    | Some (dyBarStatus _ curr_state _ bardef _) =>
      lookup_move bardef curr_state dir mv pre post
  end.

Lemma lookup_move_bp_fun: forall bp bn dir mv pre post pre' post',
  lookup_move_bp bp bn dir mv pre post ->
  lookup_move_bp bp bn dir mv pre' post' ->
  pre = pre' /\ post = post'.
Proof.
  unfold lookup_move_bp.
  intros.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  eapply lookup_move_fun; eauto.
Qed.

Lemma bpiwp_lupmvbp: forall bp bn dir mv bn' dir' mv' slt bp' pre post,
  bp_inc_waitpool bp bn dir mv slt bp' ->
  lookup_move_bp bp bn' dir' mv' pre post ->
  lookup_move_bp bp' bn' dir' mv' pre post.
Proof.
  intros.
  destruct bp.
  destruct bp'.
  destruct H.
  remember (nth_error bp_bars0 bn).
  remember (nth_error bp_bars1 bn).
  icase e.
  icase e0.
  destruct H1.
  destruct d.
  destruct d0.
  destruct dbs_wp0.
  destruct dbs_wp1.
  destruct H1 as [? [? [? [? [? [? [? ?]]]]]]].
  subst.
  case (eq_nat_dec bn bn'); intro.
  subst bn'.
  red.
  simpl.
  apply replace_nth in H2.
  rewrite H2.
  red in H0.
  simpl in *.
  rewrite <- Heqe in H0.
  trivial.
  generalize (replace_nth2 _ _ _ _ _ _ H2 n); intro.
  red.
  simpl.
  rewrite H1.
  apply H0.
Qed.

Definition lookup_slot_bp (bp : Barpool) (bn : bar_num) (dir : direction) (mv : move) (slt : slot) : Prop :=
  match nth_error (bp_bars bp) bn with
    | None => False
    | Some (dyBarStatus _ _ wp _ _) =>
       wp_dir wp = Some dir /\
       nth_error (wp_slots wp) mv = Some (Some slt)
  end.

Definition slots_unique_tid_bp (bp : Barpool) : Prop :=
  forall bn dir mv slt bn' dir' mv' slt',
  lookup_slot_bp bp bn dir mv slt ->
  lookup_slot_bp bp bn' dir' mv' slt' ->
  slot2tid slt = slot2tid slt' ->
  bn = bn' /\ dir = dir' /\ mv = mv' /\ slt = slt'.

Lemma lookup_slot_out: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall dir mv pre post,
  lookup_move_bp bp bn dir mv pre post ->
  forall slt,
  lookup_slot_bp bp bn dir mv slt ->
  exists slt',
  nth_error (dbsto_slots out) mv = Some slt' /\
  slot2tid slt' = slot2tid slt /\
  post (slot2state slt').
Proof.
  intros.
  destruct bp as [bars st wfbp].
  destruct bp' as [bars' st' wfbp'].
  destruct out as [cs dirx ns stx slts wfout].
  red in H, H0, H1.
  destruct H as [_ [_ ?]].
  unfold bp_bars in *.
  simpl.
  remember (nth_error bars bn).
  icase e.
  destruct d as [bnx csx wp bd wfdbs].
  destruct H1.
  remember (nth_error bars' bn).
  icase e.
  destruct H.
  destruct d as [bnx' csx' wp' bd' wfdbs'].
  destruct wp as [diro sltsx lim wpst wfwp].
  destruct wp' as [diro' sltsx' lim' wpst' wfwp'].
  simpl in H1, H2.
  subst diro.
  destruct H as [? [? [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]]].
  inversion H1.
  clear H1.
  subst bnx' x diro' sltsx' lim' wpst' bd' csx csx' dirx.
  clear -H16 H15 H2 H0.
  unfold lookup_move, lookup_movelist in *.
  icase (nth_error (bd_states bd) cs).
  red in H0.
  icase (nth_error (bsd_directions b) dir).
  destruct b0.
  spec H15 bml_moves.
  detach H15; trivial.
  remember (nth_error bml_moves mv).
  icase e.
  destruct p.
  destruct H0.
  subst a a0.
  red in H16.
  apply (eq_nth_error mv) in H16.
  do 3 rewrite nth_error_map in H16.
  rewrite H2 in H16.
  remember (nth_error slts mv).
  icase e.
  exists s.
  inversion H16.  
  split; trivial.  
  split; trivial.
  clear -Heqe Heqe0 H15.
  apply (assert_list_assert_holds _ _ H15 mv post (slot2state s)).
  unfold posts.
  rewrite nth_error_map.
  rewrite <- Heqe.
  trivial.
  rewrite nth_error_map.
  rewrite <- Heqe0.
  trivial.
Qed.

Definition tid_in_bp (bp : Barpool) (t : tid) : Prop :=
  exists bn, exists dir, exists mv, exists slt,
    lookup_slot_bp bp bn dir mv slt /\
    slot2tid slt = t.

Lemma lookup_slot_bp_fun: forall bp bn dir mv slt slt',
  lookup_slot_bp bp bn dir mv slt ->
  lookup_slot_bp bp bn dir mv slt' ->
  slt = slt'.
Proof.
  intros [bars st wfbp] ? ? ? ? ? ? ?.
  red in H, H0.
  simpl in H, H0.
  remember (nth_error bars bn).
  icase e.
  destruct d as [? ? wp ? ?].
  destruct H.
  destruct H0.
  congruence.
Qed.

Lemma lupslt_lupmv: forall bp bn dir mv slt,
  lookup_slot_bp bp bn dir mv slt ->
  exists pre, exists post, 
  lookup_move_bp bp bn dir mv pre post.
Proof.
  intros.
  unfold lookup_move_bp, lookup_slot_bp in *.
  remember (nth_error (bp_bars bp) bn).
  icase e.
  destruct d.
  destruct dbs_wp0.
  simpl in *.
  copy dbs_wf0.
  destruct dbs_wf1 as [? [? ?]].
  destruct H.
  subst.
  spec H2 mv.
  rewrite H3 in H2.
  destruct H2 as [pre [post [? ?]]].
  exists pre.
  exists post.
  trivial.
Qed.

Lemma bp_inc_wp_constant: forall bp bn dir mv slt bp',
  bp_inc_waitpool bp bn dir mv slt bp' ->
  forall bn' dir' mv' slt',
  (bn <> bn' \/ dir <> dir' \/ mv <> mv') ->
  lookup_slot_bp bp bn' dir' mv' slt' = lookup_slot_bp bp' bn' dir' mv' slt'.
Proof.
  intros [bars st wfbp] ? ? ? ? [bars' st' bp'] ? ? ? ? ? ?.
  destruct H.
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct H1.
  unfold lookup_slot_bp.
  simpl.
  destruct d as [bnx cs [diro slts lim stx wfwp] bd wfdbs].
  destruct d0 as [bnx' cs' [diro' slts' lim' stx' wfwp'] bd' wfdbs'].
  destruct H1 as [? [? [? [? [? [? [? ?]]]]]]].
  subst lim' cs' bnx' bd'.
  case (eq_nat_dec bn bn'); intro.
  subst bn'.
  destruct H0.
  destruct H0; trivial.
  case (eq_nat_dec dir dir'); intro.
  subst dir'.
  destruct H0; trivial.
  destruct H0; trivial.
  case (eq_nat_dec mv mv'); intro.
  subst mv'.
  destruct H0; trivial.
  rewrite <- Heqe.
  rewrite <- Heqe0.
  simpl.
  rewrite (replace_nth2 _ _ _ _ _ _ H6 n).
  destruct H4; subst; auto.
  clear -wfwp.
  destruct wfwp as [? [? ?]].
  assert (@None direction = None) by trivial.
  rewrite H0 in H2.
  clear -H2.
  apply prop_ext; split; intro; destruct H.
  disc.
  destruct (count_optionlist_nth_error _ _ H2 _ _ H0).
  rewrite <- Heqe.
  rewrite <- Heqe0.
  simpl.
  destruct H4; subst.
  apply prop_ext; split; intros [? ?]; disc.
  inversion H1.
  subst.
  contradiction n.
  trivial.
  apply prop_ext; 
  split; 
  intros [? ?]; 
  inversion H1; 
  subst; 
  contradiction n;
  trivial.
  rewrite (replace_nth2 _ _ _ _ _ _ H2 n).
  trivial.
Qed.

Lemma bpiwp_luplstbp: forall bp bn dir mv slt bp',
  bp_inc_waitpool bp bn dir mv slt bp' ->
  lookup_slot_bp bp' bn dir mv slt.
Proof.
  intros.
  destruct bp.
  destruct bp'.
  red.
  red in H.
  simpl.
  destruct H.
  icase (nth_error bp_bars0 bn).
  icase (nth_error bp_bars1 bn).
  destruct H0.
  destruct d0.
  red in H0.
  destruct d.
  destruct dbs_wp0.
  destruct dbs_wp1.
  simpl.
  destruct H0 as [? [? [? [? [? [? [? ?]]]]]]].
  subst.
  split; trivial.
  eapply replace_nth; eauto.
Qed.

(* I don't know if there is a better way but this is horrible *)
Lemma bpi_uni_nin_uni: forall bp bn dir mv slt bp',
  bp_inc_waitpool bp bn dir mv slt bp' ->
  slots_unique_tid_bp bp ->
  ~tid_in_bp bp (slot2tid slt) ->
  slots_unique_tid_bp bp'.
Proof.
  repeat intro.
  spec H0 bn0 dir0 mv0 slt0 bn'.
  spec H0 dir' mv' slt'.
  rewrite H4 in H0.
(*
  assert (bn <> bn0 \/ dir <> dir0 \/ mv <> mv0).
    case (eq_nat_dec bn bn0); auto.
    case (eq_nat_dec dir dir0); auto.
    case (eq_nat_dec mv mv0); auto.
    intros.
    subst mv0 dir0 bn0.
    generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
    apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H2.
    subst slt0.
    destruct H1.
    exists bn'.
    exists dir'.
    exists mv'.
    exists slt'.
    split; auto.
*)    
  case (eq_nat_dec bn bn0); intro.
  case (eq_nat_dec bn bn'); intro.
  case (eq_nat_dec dir dir0); intro.
  case (eq_nat_dec dir dir'); intro.
  case (eq_nat_dec mv mv0); intro.
  case (eq_nat_dec mv mv'); intro.
  do 2 subst.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H2) in H3.
  tauto.
  do 2 subst.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H2.
  subst slt0.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  destruct H1.
  exists bn'.
  exists dir'.
  exists mv'.
  exists slt'.
  split; auto.
  case (eq_nat_dec mv mv'); intro.
  do 2 subst.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H3.
  subst slt'.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  destruct H1.
  exists bn'.
  exists dir'.
  exists mv0.
  exists slt0.
  split; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  case (eq_nat_dec mv mv0); intro.
  do 2 subst.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H2.
  subst slt0.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  destruct H1.
  exists bn'.
  exists dir'.
  exists mv'.
  exists slt'.
  split; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  case (eq_nat_dec dir dir'); intro.
  case (eq_nat_dec mv mv'); intro.
  do 2 subst.  
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H3.
  subst slt'.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  destruct H1.
  exists bn'.
  exists dir0.
  exists mv0.
  exists slt0.
  split; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  case (eq_nat_dec dir dir0); intro.
  case (eq_nat_dec mv mv0); intro.
  subst.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H2.
  subst slt0.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  destruct H1.
  exists bn'.
  exists dir'.
  exists mv'.
  exists slt'.
  split; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  case (eq_nat_dec bn bn'); intro.
  case (eq_nat_dec dir dir'); intro.
  case (eq_nat_dec mv mv'); intro.
  subst.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H3.
  subst slt'.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  destruct H1.
  exists bn0.
  exists dir0.
  exists mv0.
  exists slt0.
  split; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H2; auto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H) in H3; auto.
Qed.

Lemma bp_trans_out2: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall mv slt,
  nth_error (dbsto_slots out) mv = Some slt ->
  exists slt',
  lookup_slot_bp bp bn (dbsto_dir out) mv slt' /\
  slot2tid slt = slot2tid slt'.
Proof.
  intros [bars st wfbp] ? [bars' st' bp'] [cs dir ds st'' slts ?] [? [? ?]] ? ? ?.
  unfold lookup_slot_bp.
  unfold bp_bars.
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct H1.
  unfold dbsto_slots in H2.
  destruct d as [bn' cs' [diro slts' lim st''' ?] bd wfdbs].
  destruct d0 as [bn'' cs'' [diro' slts'' lim' st'''' ?] bd' wfdbs'].
  simpl.
  red in H1.
  destruct H1 as [? [dirx [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]]].
  subst.
  clear -H2 H17.
(* better way? *)
  red in H17.
  revert slts' mv slt H17 H2.
  induction slts; intros.
  icase mv.
  destruct mv.
  inversion H2.
  subst.
  simpl in H17.
  icase slts'.
  simpl in H17.
  icase o.
  inversion H17.
  exists s.
  split; trivial.
  split; trivial.
  simpl in H2.
  icase slts'.
  inversion H17.
  simpl.
  auto.
(* end better way? *)
Qed.

Lemma bptrans_slt: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall bn',
  bn <> bn' ->
  forall dir mv slt,
  lookup_slot_bp bp bn' dir mv slt ->
  lookup_slot_bp bp' bn' dir mv slt.
Proof.
  unfold bp_transition, lookup_slot_bp.
  intros [bars st wfbp] ? [bars' st' wfbp'] [? ? ? ? ? ?] ? ? ? ? ? ? ?.
  destruct H as [_ [_ ?]].
  unfold bp_bars in *.
  icase (nth_error bars bn).
  icase (nth_error bars' bn).
  destruct H.
  rewrite (replace_nth2 _ _ _ _ _ _ H2 H0).
  trivial.
Qed.

Lemma bptrans_out_tid: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall tid,
  In tid (map slot2tid (dbsto_slots out)) ->
  exists dir, exists mv, exists slt,
    lookup_slot_bp bp bn dir mv slt /\
    tid = slot2tid slt.
Proof.
  intros [bars st wfbp] ? [bars' st' wfbp'] [? ? ? ? ? ?] ? ? ?.
  unfold lookup_slot_bp.
  simpl in *.
  destruct H as [_ [_ ?]].
  icase (nth_error bars bn).
  icase (nth_error bars' bn).
  destruct H.
  destruct d.
  destruct d0.
  destruct dbs_wp0.
  destruct dbs_wp1.
  simpl in *.
  red in H.
  destruct H as [? [dir [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]]].
  subst.
  exists dir.
  clear -H0 H15.
  revert wp_slots0 H0 H15.
  induction dbsto_slots0; intros.
  contr.
  simpl in *.
  icase wp_slots0.
  icase o.
  destruct H0.
  exists 0.
  exists s.
  simpl.
  red in H15.
  simpl in H15.
  inversion H15.
  rewrite H1.
  rewrite H.
  tauto.
  inversion H15.
  spec IHdbsto_slots0 wp_slots0 H H2.
  destruct IHdbsto_slots0 as [mv [slt [[? ?] ?]]].
  exists (S mv).
  exists slt.
  simpl.
  tauto.
Qed.

Lemma bptrans_nio: forall bp bn bp' out,
  slots_unique_tid_bp bp ->
  bp_transition bp bn bp' out ->
  forall bn' dir mv slt,
  lookup_slot_bp bp bn' dir mv slt ->
  In (slot2tid slt) (map slot2tid (dbsto_slots out)) ->
  bn = bn'.
Proof.
  intros.
  apply (bptrans_out_tid _ _ _ _ H0) in H2.
  destruct H2 as [dir' [mv' [slt' [? ?]]]].
  spec H bn dir' mv' slt' bn'.
  spec H dir mv slt H2 H1.
  symmetry in H3.
  spec H H3.
  tauto.
Qed.

Definition lookup_bstate_bp (bp : Barpool) (bn : bar_num) (cst : stateid) : Prop :=
  match nth_error (bp_bars bp) bn with
    | None => False
    | Some (dyBarStatus _ curr_state _ _ _) =>
      cst = curr_state
  end.

Lemma lookup_move_bp_lookup_state_bp: forall bp bn dir mv pre post,
  lookup_move_bp bp bn dir mv pre post ->
  exists cst,
    lookup_bstate_bp bp bn cst.
Proof.
  unfold lookup_move_bp, lookup_bstate_bp.
  intros.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  exists dbs_cs0.
  trivial.
Qed.

Lemma bpiwp_lupbst: forall bp bn dir mv slt bp',
  bp_inc_waitpool bp bn dir mv slt bp' ->
  forall cst bn',
  lookup_bstate_bp bp bn' cst ->
  lookup_bstate_bp bp' bn' cst.
Proof.
  unfold bp_inc_waitpool, lookup_bstate_bp.
  intros [bars st wfbp] ? ? ? ? [bars' st' wfbp'] ? ? ? ?.
  destruct H.
  simpl in *.
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct H1.
  remember (nth_error bars bn').
  icase e.
  case (eq_nat_dec bn bn'); intro.
  subst bn'.
  rewrite <- Heqe in Heqe1.
  inversion Heqe1.
  subst d1.
  clear Heqe1.
  apply replace_nth in H2.
  rewrite H2.
  destruct d.
  destruct d0.
  subst.
  destruct dbs_wp0.
  destruct dbs_wp1.
  destruct H1 as [_ [? _]].
  trivial.
  rewrite (replace_nth2 _ _ _ _ _ _ H2 n).
  rewrite <- Heqe1.
  trivial.
Qed.

Lemma bptrans_bstate: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall bn',
  bn <> bn' ->
  forall cst,
  lookup_bstate_bp bp bn' cst ->
  lookup_bstate_bp bp' bn' cst.
Proof.
  unfold bp_transition, lookup_bstate_bp.
  intros [bars st wfbp] ? [bars' st' wfbp'] [? ? ? ? ? ?] ? ? ? ? ?.
  destruct H as [_ [_ ?]].
  unfold bp_bars in *.
  icase (nth_error bars bn).
  icase (nth_error bars' bn).
  destruct H.
  rewrite (replace_nth2 _ _ _ _ _ _ H2 H0).
  trivial.
Qed.

Lemma bptrans_lupmvbp_lupbst: forall bp bn bp' out bn' dir mv pre post,
  bp_transition bp bn bp' out ->
  lookup_move_bp bp bn' dir mv pre post ->
  exists cst, lookup_bstate_bp bp' bn' cst.
Proof.
  intros.
  destruct bp.
  destruct bp'.
  destruct out.
  destruct H as [? [? ?]].
  red in H0.
  unfold lookup_bstate_bp.
  simpl in *.
  remember (nth_error bp_bars0 bn).
  remember (nth_error bp_bars1 bn).
  icase e.
  icase e0.
  destruct H2.
  case (eq_nat_dec bn bn'); intro.
  subst bn'.
  rewrite <- Heqe0.
  destruct d0.
  exists dbs_cs0.
  trivial.
  rewrite (replace_nth2 _ _ _ _ _ _ H3 n).
  icase (nth_error bp_bars0 bn').
  destruct d1.
  exists dbs_cs0.
  trivial.
Qed.

Lemma bp_trans_st: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall bs bs',
  lookup_bstate_bp bp bn bs ->
  lookup_bstate_bp bp' bn bs' ->
  (sepconj Tru (barrier_assert bn pfullshare bs)) (bp_st bp) /\
  join (bp_st bp') (dbsto_st out) (st_upd_bm_bn (bp_st bp) bn (barst_builder bs' pfullshare)).
Proof.
  intros [bars bst wfbp] ? [bars' bst' wfbp'] [ss dir ds ost oslts wfo].
  intros.
  unfold bp_st, dbsto_st.
  red in H, H0, H1.
  unfold bp_bars in *.
  destruct H as [? [? ?]].
  remember (nth_error bars bn).
  icase e.
  remember (nth_error bars' bn).
  icase e.
  destruct d.
  destruct d0.
  subst dbs_cs0 dbs_cs1.
  destruct dbs_wp0.
  destruct dbs_wp1.
  destruct H3.
  destruct H0 as [? [dirx [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]]].
  subst dbs_bn0 wp_dir0 wp_dir1 wp_slots1 wp_limit0 wp_st1 dbs_bd0 dirx ss ds ost.
  destruct bst as [[? ?] ?].
  destruct bst' as [[? ?] ?].
  unfold st_bsm, snd  in H.
  split.
  exists (s, h, upd_bm b bn None).
  exists (s, empty_heap, fun n => if eq_nat_dec n bn then barst_builder bs pfullshare else None).
  split.
  split.
  split.
  split; trivial.
  apply join_com.
  apply join_emptyheap.
  unfold upd_bm.
  unfold snd.
  intro.
  case (eq_nat_dec x bn); intro.
  subst x.
  rewrite H.
  constructor.
  constructor.
  split.
  exact I.
  split.
  unfold st_bsm, snd.
  case (eq_nat_dec bn bn); intro hx; auto.
  destruct hx; trivial.
  split.
  intros.
  unfold st_bsm, snd.
  case (eq_nat_dec bn' bn); intro hx; contr; auto.
  trivial.
  apply join_com.
  trivial.
Qed.

Lemma bpiwp_st: forall bp bn dir mv slt bp',
  bp_inc_waitpool bp bn dir mv slt bp' ->
  join (bp_st bp) (slot2state slt) (bp_st bp').
Proof.
  intros.
  destruct bp.
  destruct bp'.
  destruct H.
  apply join_com.
  trivial.
Qed.

Definition lookup_state_bp (bp : Barpool) (bn : bar_num) (st : state) : Prop :=
  match nth_error (bp_bars bp) bn with
    | None => False
    | Some (dyBarStatus _ _ (waitpool _ _ _ st' _) _ _) =>
      st = st'
  end.

Lemma bpiwp_state: forall bp bn st,
   lookup_state_bp bp bn st ->
   forall slt st',
   join st (slot2state slt) st' ->
   forall dir mv bp',
   bp_inc_waitpool bp bn dir mv slt bp' ->
   lookup_state_bp bp' bn st'.
Proof.
  intros [bars st wfbp] ? ? ? ? ? ? ? ? [bars' st' wfbp'] ?.
  destruct H1.
  red in H.
  red.
  unfold bp_bars in *.
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct d.
  destruct d0.
  destruct dbs_wp0.
  destruct dbs_wp1.
  subst wp_st0.
  destruct H2.
  destruct H as [? [? [? [? [? [? [? ?]]]]]]].
  subst.
  apply join_com in H8.
  apply (join_eq _ _ _ _ H0) in H8.
  trivial.
Qed.

Lemma lupmvbp_lupstbp: forall bp bn dir mv pre post,
  lookup_move_bp bp bn dir mv pre post ->
  exists st,
    lookup_state_bp bp bn st.
Proof.
  intros.
  unfold lookup_move_bp, lookup_state_bp in *.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  destruct dbs_wp0.
  exists wp_st0.
  trivial.
Qed.

Lemma lookup_move_bp_facts: forall bp bn dir mv pre post,
  lookup_move_bp bp bn dir mv pre post ->
  exists cst,
  lookup_bstate_bp bp bn cst /\
  ubiquitous pre /\
  ubiquitous post /\
  precise pre /\
  (exists ps : pshare,
    derives pre (sepconj (barrier_assert bn ps cst) Tru)).
Proof.
  intros.
  red in H.
  unfold lookup_bstate_bp.
  remember (nth_error (bp_bars bp) bn).
  icase e.
  destruct d.
  exists dbs_cs0.
  apply lookup_move_facts in H.
  destruct H as [? [? [? [ps ?]]]].
  repeat (split; trivial).
  exists ps.
  destruct bp.
  simpl in Heqe.
  destruct bp_wf0.
  symmetry in Heqe.
  apply H4 in Heqe.
  simpl in Heqe.
  subst dbs_bn0.
  red in dbs_wf0.
  destruct dbs_wp0.
  destruct dbs_wf0 as [? [? ?]].
  subst.
  trivial.
Qed.

Definition lookup_dir_bp (bp : Barpool) (bn : bar_num) (dir : direction) : Prop :=
  match nth_error (bp_bars bp) bn with
    | None => False
    | Some (dyBarStatus _ _ (waitpool None _ _ _ _) _ _) => False
    | Some (dyBarStatus _ _ (waitpool (Some dir') _ _ _ _) _ _) =>
      dir = dir'
  end.

Lemma bp_ready_lupmv_lupslt: forall bp bn,
  bp_ready bp bn ->
  forall dir mv pre post,
  lookup_move_bp bp bn dir mv pre post ->
  lookup_dir_bp bp bn dir ->
  exists slt,
  lookup_slot_bp bp bn dir mv slt.
Proof.
  intros.
  rename H1 into dirx.
  unfold lookup_dir_bp, lookup_move_bp, bp_ready, lookup_slot_bp in *.
  icase (nth_error (bp_bars bp) bn).
  destruct H as [? [? ?]].
  inversion H.
  subst x.
  clear H.
  destruct d.
  unfold dbs_ready, lookup_move in *.
  simpl in *.
  remember (nth_error (bd_states dbs_bd0) dbs_cs0).
  icase e.
  destruct dbs_wp0 as [diro slts lim st wpwf].
  simpl.
  destruct dbs_wf0 as [? [? ?]].
  destruct dbs_bd0 as [bn' lim' sts].
  simpl in *.
  subst lim' dbs_bn0.
  destruct wpwf as [? [? ?]].
  subst lim.
  copy bd_wf.
  destruct bd_wf0 as [? ?].
  spec H5 dbs_cs0 b.
  rewrite Heqe in H5.
  detach H5; trivial.
  destruct H5 as [? [? ?]].
  subst.
  destruct diro; contr.
  subst d.
  spec H3 mv.
  remember (nth_error slts mv).
  destruct e.
  (* case 1 *)
  destruct o.
  2: symmetry in Heqe0; elimtype False; eapply fullo_no_None; eauto.
  exists s.
  split; trivial.
  
  (* case 2 *)
  symmetry in Heqe0.
  apply nth_error_bound2 in Heqe0.
  rewrite <- H5 in Heqe0.
  destruct b.
  simpl in *.
  copy bsd_wf.
  destruct bsd_wf0 as [_ ?].
  red in H0.
  remember (nth_error bsd_directions dir).
  icase e.
  spec H6 dir b.
  detach H6; auto.
  destruct H6 as [? [? ?]].
  subst.
  destruct b.
  simpl in *.
  copy bml_wf.
  destruct bml_wf0 as [_ [? _]].
  remember (nth_error bml_moves mv).
  icase e.
  symmetry in Heqe2.
  apply nth_error_bound in Heqe2.
  omegac.
Qed.

Lemma bp_lupmvslt_pre: forall bp bn dir mv pre post slt,
  lookup_move_bp bp bn dir mv pre post ->
  lookup_slot_bp bp bn dir mv slt ->
  pre (slot2state slt).
Proof.
  intros [bars st bpwf] ? ? ? ? ? ? ? ?.
  red in H, H0.
  simpl in *.
  icase (nth_error bars bn).
  destruct d as [bn' cs wp bd dbswf].
  destruct H0.
  destruct wp as [diro slts lim st' wpwf].
  simpl in *.
  subst diro.
  destruct dbswf as [? [? ?]].
  subst lim bn'.
  spec H3 mv.
  rewrite H1 in H3.
  destruct H3 as [pre' [post' [? ?]]].
  apply (lookup_move_fun _ _ _ _ _ _ _ _ H) in H0.
  destruct H0.
  subst.
  trivial.
Qed.

Lemma bp_lupsltst: forall bp bn st dir mv slt,
  lookup_state_bp bp bn st ->
  lookup_slot_bp bp bn dir mv slt ->
  join_sub (slot2state slt) st.
Proof.
  intros.
  red in H, H0.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  destruct dbs_wp0.
  simpl in *.
  destruct H0.
  subst.
  destruct wp_wf0 as [? [? ?]].
  clear -H1 H2.
  apply wp_getstates_SomeSome in H1.
  assert (join empty_state (slot2state slt) (slot2state slt)).
    split.
    2: apply join_emptybm.
    split.
    2: apply join_emptyheap.
    destruct slt as [? [? ?]].
    split; reflexivity.
  destruct (join_list_split_nth _ _ _ H1 H2 H) as [? [sts' [? [? ?]]]].
  apply empty_state_identity in H3.
  subst x.
  apply join_com in H.
  destruct (join_list_assoc1 H H4) as [st' [? ?]].
  exists st'.
  trivial.
Qed.

(* I also need this somewhat odd varient... *)
Lemma bp_lupsltst2: forall bp bn st dir mv slt mv' slt',
  lookup_state_bp bp bn st ->
  lookup_slot_bp bp bn dir mv slt ->
  lookup_slot_bp bp bn dir mv' slt' ->
  mv <> mv' ->
  exists st',
  join (slot2state slt) (slot2state slt') st' /\
  join_sub st' st.
Proof.
  intros.
  red in H, H0, H1.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  destruct dbs_wp0.
  simpl in *.
  destruct H0.
  destruct H1.
  subst.
  destruct wp_wf0 as [? [? ?]].
  clear -H2 H3 H4 H5.
  apply wp_getstates_SomeSome in H3.
  apply wp_getstates_SomeSome in H4.
  assert (join empty_state (slot2state slt) (slot2state slt)).
    apply join_emptystate.
    destruct slt as [? [? ?]].
    trivial.
  destruct (join_list_split_nth _ _ _ H3 H5 H) as [? [sts' [? [? ?]]]].
  apply empty_state_identity in H1.
  subst x.
  assert (join empty_state (slot2state slt') (slot2state slt')).
    apply join_emptystate.
    destruct slt' as [? [? ?]].
    trivial.
  generalize (replace_nth2 _ _ _ _ _ _ H0 H2); intro.
  rewrite H4 in H7.
  destruct (join_list_split_nth _ _ _ H7 H6 H1) as [? [sts'' [? [? ?]]]].
  assert (join x empty_state x).
    apply join_com.
    apply join_emptystate.
    destruct x as [[? ?] ?].
    destruct H9 as [[[_ ?] _] _].
    destruct slt' as [? [? ?]].
    simpl in H9.
    subst; trivial.
  destruct (join_list_assoc1 H11 H10) as [st' [? ?]].
  exists x.
  split; trivial.
  exists st'.
  trivial.
Qed.

Lemma lupslt_dir: forall bp bn dir mv slt,
  lookup_slot_bp bp bn dir mv slt ->
  lookup_dir_bp bp bn dir.
Proof.
  unfold lookup_slot_bp, lookup_dir_bp.
  intros.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  destruct dbs_wp0.
  destruct H.
  simpl in *.
  subst.
  trivial.
Qed.

Lemma bptrans_slt2: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall bn',
  bn <> bn' ->
  forall dir mv slt,
  lookup_slot_bp bp' bn' dir mv slt ->
  lookup_slot_bp bp bn' dir mv slt.
Proof.
  unfold bp_transition, lookup_slot_bp.
  intros [bars st wfbp] ? [bars' st' wfbp'] [? ? ? ? ? ?] ? ? ? ? ? ? ?.
  destruct H as [_ [_ ?]].
  unfold bp_bars in *.
  icase (nth_error bars bn).
  icase (nth_error bars' bn).
  destruct H.
  rewrite <- (replace_nth2 _ _ _ _ _ _ H2 H0).
  trivial.
Qed.

Lemma bptrans_bpnin: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  forall dir mv slt,
  lookup_slot_bp bp' bn dir mv slt ->
  False.
Proof.
  intros [bars bpst wfbp] ? [bars' bpst' wfbp'] [? ? ? ? ? ?] ? ? ? ? ?.
  destruct H as [_ [_ ?]].
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct H.
  red in H0.
  simpl in H0.
  rewrite <- Heqe0 in H0.
  destruct d.
  destruct dbs_wp0.
  destruct d0.
  destruct dbs_wp0.
  destruct H0.
  destruct H as [? [? [? [? ?]]]].
  simpl in *.
  subst.
  disc.
Qed.

(* This is awful but I am too tired to it to improve it right now... *)
Lemma stuck_not_bp_ready: forall mv mv' bp bn dir pre post pre' post',
  bp_ready bp bn ->
  lookup_move_bp bp bn dir mv pre post ->
  lookup_move_bp bp bn dir mv' pre' post' ->
  forall st,
  lookup_state_bp bp bn st ->
  forall slt,
  lookup_slot_bp bp bn dir mv slt ->
  forall stx stx',
  join stx (slot2state slt) stx' ->
  joins stx st ->
  (sepconj pre' Tru) stx' ->
  mv = mv'.
Proof.
  intros.
  destruct (bp_ready_lupmv_lupslt _ _ H _ _ _ _ H1) as [slt' ?].
  eapply lupslt_dir; eauto.
  generalize (bp_lupmvslt_pre _ _ _ _ _ _ _ H1 H7); intro.
  generalize (bp_lupsltst _ _ _ _ _ _ H2 H7); intro.
  generalize (bp_lupsltst _ _ _ _ _ _ H2 H3); intro.
  destruct (lookup_move_bp_facts _ _ _ _ _ _ H1) as [cst [_ [_ [_ [? ne]]]]].
  destruct stx' as [[? ?] ?].
  inv_sc H6.
  subst s0 s1.
  clear H16.
  destruct H5 as [stx' ?].
  spec H11 stx' (slot2state slt') (s, h0, b0).
  detach H11.
  2: assert (join_sub st stx') by (exists stx; apply join_com; trivial); eapply join_sub_trans; eauto.
  detach H11.
  2: assert (join_sub (s, h0, b0) (s, h,b)) by (exists (s, h1, b1); repeat (split; trivial)); apply join_sub_trans with (s, h, b); auto.
  2: destruct H10 as [? ?]; apply join_com in H5; apply join_com in H10.
  2: destruct (join_assoc _ _ _ _ _ H10 H5) as [? [? ?]]; apply join_com in H4; apply (join_eq _ _ _ _ H4) in H11; subst x0.
  2: exists x; apply join_com; trivial.
  spec H11 H8 H15.
  destruct slt' as [tidx [? ?]].
  inversion H11.
  subst s h2 b2.
  clear H11 H8.
  case (eq_nat_dec mv mv'); trivial; intro.
  destruct (bp_lupsltst2 _ _ _ _ _ _ _ _ H2 H3 H7 n) as [slts_st [? ?]].
  destruct H8 as [oslts ?].
  apply join_com in H5.
  apply join_com in H8.
  destruct (join_assoc _ _ _ _ _ H8 H5) as [st0 [? ?]].
  apply join_com in H6.
  destruct (join_assoc _ _ _ _ _ H6 H11) as [st1 [? ?]].
  apply join_com in H16.
  apply (join_eq _ _ _ _ H4) in H16.
  subst st1.
  assert (join (empty_stack, h0, b0) (empty_stack, h1,b1) (empty_stack, h, b)) by (repeat (split; trivial)).
  apply join_com in H17.
  apply join_com in H16.
  destruct (join_assoc _ _ _ _ _ H16 H17) as [e [? ?]].
  generalize (join_self _ _ H18); intro.
  subst e.
  assert (identity (empty_stack, h0, b0)) by (rewrite identity_unit_equiv; trivial).
  generalize (identity_emptystate _ H20); intro.
  detach H21; trivial.
  inversion H21.
  subst h0 b0.
  destruct ne.
  apply H22 in H15.
  clear -H15.
  inv_sc H15.
  subst s s0.
  clear H4.
  generalize (split_identity _ _ H2 empty_bm_identity); intro.
  apply identity_emptybm in H.
  subst b.
  destruct H3 as [? [? ?]].
  destruct x.
  inversion H.
Qed.

Definition bp_bardefs (bp : Barpool) : list BarDef :=
  map dbs_bd (bp_bars bp).

Lemma barDefs_bp_inc_waitpool_preserved: forall bp bn dir mv slt bp',
  bp_inc_waitpool bp bn dir mv slt bp' ->
  bp_bardefs bp = bp_bardefs bp'.
Proof.
  intros.
  destruct bp.
  destruct bp'.
  red in H.
  destruct H.
  remember (nth_error bp_bars0 bn).
  remember (nth_error bp_bars1 bn).
  icase e.
  icase e0.
  destruct H0.
  red in H0.
  destruct d.
  destruct dbs_wp0.
  destruct d0.
  destruct dbs_wp0.
  destruct H0 as [? [? [? [? [? [? [? ?]]]]]]].
  subst.
  unfold bp_bardefs.
  simpl.
  remember (waitpool wp_dir0 wp_slots0 wp_limit1 wp_st0 wp_wf0).
  remember (waitpool (Some dir) wp_slots1 wp_limit1 wp_st1 wp_wf1).
  clear -Heqe H1 Heqe0.
  apply nth_error_eq.
  intro.
  do 2 rewrite nth_error_map.
  case (eq_nat_dec bn i); intro.
  subst i.
  rewrite <- Heqe.
  rewrite <- Heqe0.
  simpl.
  trivial.
  rewrite (replace_nth2 _ _ _ _ _ _ H1 n).
  trivial.
Qed.

Lemma barDefs_bp_transition_preserved: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  bp_bardefs bp = bp_bardefs bp'.
Proof.
  intros.
  destruct bp.
  destruct bp'.
  destruct out.
  red in H.
  destruct H as [? [? ?]].
  remember (nth_error bp_bars0 bn).
  remember (nth_error bp_bars1 bn).
  icase e.
  icase e0.
  destruct H1.
  red in H1.
  destruct d.
  destruct dbs_wp0.
  destruct d0.
  destruct dbs_wp0.
  destruct H1 as [? [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]].
  subst.
  remember (waitpool (Some x) wp_slots0 wp_limit1 wp_st0 wp_wf0).
  remember (waitpool None (n_None slot wp_limit1) wp_limit1 empty_state wp_wf1).
  clear -Heqe Heqe0 H2.
  apply nth_error_eq.
  intro.
  unfold bp_bardefs.
  do 2 rewrite nth_error_map.
  simpl.
  case (eq_nat_dec bn i); intro.
  subst i.
  rewrite <- Heqe.
  rewrite <- Heqe0.
  simpl.
  trivial.
  rewrite (replace_nth2 _ _ _ _ _ _ H2 n).
  trivial.
Qed.

Definition sched : Type := list nat.

Inductive thd_state : Type :=
  | Running : cmd -> thd_state
  | Waiting : bar_num -> direction -> move -> cmd -> thd_state.

Record Thread : Type := thread {
  th_stk : stack;
  th_hp : heap;
  th_bs : BarstMap;
  th_ctl : thd_state}.

Definition tl_states (tl : list Thread) : list state :=
  map (fun t => (empty_stack, th_hp t, th_bs t)) tl.

Lemma tl_states_lookup: forall thds i stk h bs ctl,
  nth_error thds i  = Some (thread stk h bs ctl) ->
  nth_error (tl_states thds) i = Some (empty_stack, h, bs).
Proof.
  intros.
  unfold tl_states.
  rewrite nth_error_map.
  rewrite H.
  trivial.
Qed.

Fixpoint transition_threads' (slots : list slot) (threads : list Thread) 
                                               (threads' : list Thread) : Prop :=
  match slots with
    | nil => threads = threads'
    | (tn, (h, bs)) :: slots' => 
       match (nth_error threads tn) with
         | None => False
         | Some (thread stk h' bs' ctl) =>
               exists h'', join h h' h'' /\
               exists bs'', join bs bs' bs'' /\
               exists bn, exists dir, exists mv, exists k,
               ctl = Waiting bn dir mv k /\
               match (replace threads tn (thread stk h'' bs'' (Running k))) with
                 | None => False
                 | Some threads2 =>
                       transition_threads' slots' threads2 threads'
               end
       end
  end.

Definition transition_threads (out : DbsTransOut) (threads : list Thread) (threads' : list Thread) : Prop :=
  transition_threads' (dbsto_slots out) threads threads'.

Lemma transition_threads'_constant: forall slts th th',
  transition_threads' slts th th' ->
  forall tid,
  ~In tid (map slot2tid slts) ->
  nth_error th tid = nth_error th' tid.
Proof.
  induction slts; intros.
  simpl in H.
  subst.
  trivial.
  destruct a as [? [? ?]].
  simpl in H, H0.
  remember (nth_error th t).
  icase e.
  destruct t0 as [stk h' bs' ctl].
  destruct H as [h'' [? [bs'' [? ?]]]].
  destruct H2 as [bn [dir [mv [k [? ?]]]]].
  remember (replace th t (thread stk h'' bs'' (Running k))).
  icase o.
  spec IHslts l th' H3 tid0.
  assert (t <> tid0 /\ ~ In tid0 (map slot2tid slts)) by (split; intro; subst; auto).
  destruct H4.
  spec IHslts H5.
  symmetry in Heqo.
  rewrite <- (replace_nth2 _ _ _ _ _ _ Heqo H4).
  trivial.
Qed.

Definition slts_unique_tids (slts : list slot) : Prop :=
  forall i j slt slt',
    nth_error slts i = Some slt ->
    nth_error slts j = Some slt' ->
    slot2tid slt = slot2tid slt' ->
    i = j.

Lemma bpt_uni_out: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  slots_unique_tid_bp bp ->
  slts_unique_tids (dbsto_slots out).
Proof.
  repeat intro.
  destruct bp.
  destruct bp'.
  destruct out.
  destruct H as [? [? ?]].
  simpl in *.
  red in H0.
  remember (nth_error bp_bars0 bn).
  remember (nth_error bp_bars1 bn).
  icase e.
  icase e0.
  destruct H5.
  red in H5.
  destruct d.
  destruct d0.
  destruct dbs_wp0.
  destruct dbs_wp1.
  simpl in *.
  destruct H5 as [? [dir [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]]].
  subst.
  unfold lookup_slot_bp in H0.
  simpl in H0.
  unfold dualolist in H20.
  assert (exists slta, nth_error wp_slots0 i = Some (Some slta) /\ slot2tid slt = slot2tid slta).
    clear -H1 H20.
    apply (eq_nth_error i) in H20.
    do 3 rewrite nth_error_map in H20.
    rewrite H1 in H20.
    simpl in *.
    icase (nth_error wp_slots0 i).
    icase o.
    exists s.
    inversion H20.
    split; auto.
  destruct H5 as [slt1 [? ?]].
  assert (exists slta, nth_error wp_slots0 j = Some (Some slta) /\ slot2tid slt' = slot2tid slta).
    clear -H2 H20.
    apply (eq_nth_error j) in H20.
    do 3 rewrite nth_error_map in H20.
    rewrite H2 in H20.
    simpl in *.
    icase (nth_error wp_slots0 j).
    icase o.
    exists s.
    inversion H20.
    split; auto.
  destruct H8 as [slt2 [? ?]].
  spec H0 bn dir i slt1.
  spec H0 bn dir j slt2.
  rewrite <- Heqe in H0.
  simpl in H0.
  rewrite H5, H8 in H0.
  destruct H0 as [? [? [? ?]]]; auto.
  congruence.
Qed.

Lemma slts_unique_head: forall t h b slts,
  slts_unique_tids ((t,(h,b)) :: slts) ->
  slts_unique_tids slts.
Proof.
  repeat intro.
  spec H (S i) (S j) slt slt'.
  spec H H0 H1 H2.
  inversion H.
  trivial.
Qed.

Lemma slts_unique_head_notin: forall t h b slts,
  slts_unique_tids ((t,(h,b)) :: slts) ->
  ~In t (map slot2tid slts).
Proof.
  induction slts; intros.
  intro.
  contr.
  simpl.
  intro.
  destruct H0.
  subst t.
  spec H 0 1 (slot2tid a, (h, b)) a.
  detach H; auto.
  detach H; auto.
  detach H; auto.
  inversion H.
  apply IHslts; auto.
  repeat intro.
  destruct i.
  destruct j; auto.
  spec H 0 (S (S j)) slt slt'.
  spec H H1 H2 H3.
  inversion H.
  destruct j.
  spec H (S (S i)) 0 slt slt'.
  spec H H1 H2 H3.
  disc.
  simpl in H1, H2.
  spec H (S (S i)) (S (S j)) slt slt'.
  spec H H1 H2 H3.
  inversion H.
  trivial.
Qed.

Lemma transition_threads'_changes: forall slts th th',
  slts_unique_tids slts ->
  transition_threads' slts th th' ->
  forall i slt,
  nth_error slts i = Some slt ->
  forall stk hp bs bn dir mv k,
  nth_error th (slot2tid slt) = Some (thread stk hp bs (Waiting bn dir mv k)) ->
  exists hx, exists bx,
  join (slot2state slt) (empty_stack, hp, bs) (empty_stack, hx, bx) /\
  nth_error th' (slot2tid slt) = Some (thread stk hx bx (Running k)).
Proof.
  induction slts; intros.
  icase i.
  destruct i.
  inversion H1.
  subst a.
  clear H1.
  destruct slt as [? [? ?]].
  simpl in H0.
  remember (nth_error th t).
  icase e.
  destruct t0.
  destruct H0 as [hx [? [bx [? [bnx [dirx [mvx [kx [? ?]]]]]]]]].
  subst.
  simpl in H2.
  rewrite H2 in Heqe.
  inversion Heqe.
  subst th_stk0 th_hp0 th_bs0 bnx dirx mvx kx.
  clear Heqe.
  remember (replace th t (thread stk hx bx (Running k))).
  icase o.
  exists hx.
  exists bx.
  repeat (split; trivial).
  simpl.
  rewrite <- (transition_threads'_constant _ _ _ H4).
  eapply replace_nth; eauto.
  eapply slts_unique_head_notin; eauto.
  destruct a as [? [? ?]].
  simpl in H1, H0.
  remember (nth_error th t).
  icase e.
  destruct t0.
  destruct H0 as [hx [? [bx [? [bnx [dirx [mvx [kx [? ?]]]]]]]]].
  remember (replace th t (thread th_stk0 hx bx (Running kx))).
  icase o.
  copy H.
  apply slts_unique_head in H.
  spec IHslts l th' H H5.
  spec IHslts i slt H1.
  spec IHslts stk hp bs bn.
  spec IHslts dir mv k.
  apply IHslts.
  clear IHslts.
  case (eq_nat_dec t (slot2tid slt)).
  intro.
  subst t.
  spec H6 0 (S i) (slot2tid slt, (h, b)) slt.
  detach H6; auto.
  detach H6; auto.
  detach H6; auto.
  inversion H6.
  intro.
  symmetry in Heqo.
  rewrite (replace_nth2 _ _ _ _ _ _ Heqo n).
  trivial.
Qed.

Lemma transition_threads_constant: forall out th th',
  transition_threads out th th' ->
  forall tid,
  ~In tid (map slot2tid (dbsto_slots out)) ->
  nth_error th tid = nth_error th' tid.
Proof.
  intro.
  destruct out as [? ? ? ? slts ?].
  unfold transition_threads.
  simpl.
  clear.
  eapply transition_threads'_constant; eauto.
Qed.

Lemma transition_threads_changes: forall out th th',
  transition_threads out th th' ->
  slts_unique_tids (dbsto_slots out) ->
  forall i slt,
  nth_error (dbsto_slots out) i = Some slt ->
  forall stk hp bs bn dir mv k,
  nth_error th (slot2tid slt) = Some (thread stk hp bs (Waiting bn dir mv k)) ->
  exists hx, exists bx,
  join (slot2state slt) (empty_stack, hp, bs) (empty_stack, hx, bx) /\
  nth_error th' (slot2tid slt) = Some (thread stk hx bx (Running k)).
Proof.
  intro.
  destruct out as [? ? ? ? slts ?].
  unfold transition_threads.
  simpl.
  clear.
  intros.
  generalize (transition_threads'_changes _ _ _ H0 H _ _ H1 _ _ _ _ _ _ _ H2).
  auto.
Qed.

Lemma transition_threads'_fun: forall out thds thds' thds'',
  transition_threads' out thds thds' ->
  transition_threads' out thds thds'' ->
  thds' = thds''.
Proof.
  induction out; intros.
  inversion H.
  inversion H0.
  congruence.
  simpl in H, H0.
  destruct a as [? [? ?]].
  icase (nth_error thds t).
  destruct t0.
  destruct H as [hx [? [bx [? ?]]]].
  destruct H0 as [hx' [? [bx' [? ?]]]].
  assert (join h th_hp0 hx) by trivial.
  assert (join h th_hp0 hx') by trivial.
  apply (join_eq _ _ _ _ H5) in H6.
  subst hx'.
  assert (join b th_bs0 bx) by trivial.
  assert (join b th_bs0 bx') by trivial.
  apply (join_eq _ _ _ _ H6) in H7.
  subst bx'.
  destruct H4 as [bn [dir [mv [k [? ?]]]]].
  destruct H2 as [bn' [dir' [mv' [k' [? ?]]]]].
  rewrite H2 in H4.
  inversion H4.
  subst bn' dir' mv' k'.
  icase (replace thds t (thread th_stk0 hx bx (Running k))).
  eapply IHout; eauto.
Qed.

Lemma can_transition_threads': forall out,
  slts_unique_tids out ->
  forall outst,
  join_list empty_state (map slot2state out) outst ->
  forall thds thdst,
  join_list empty_state (tl_states thds) thdst ->
  forall tst,
  join outst thdst tst ->
  (forall i, 
    In i (map slot2tid out) -> 
      match nth_error thds i with
        | Some (thread stk hp bs (Waiting bn dir mv c)) => True | _ => False end) ->
  exists thds',
  transition_threads' out thds thds'.
Proof.
  induction out; intros.
  exists thds.
  reflexivity.
  destruct a as [? [? ?]].
  spec IHout (slts_unique_head _ _ _ _ H).
  unfold transition_threads'.
  fold transition_threads'.
  assert (map slot2state ((t, (h, b)) :: out) = (empty_stack, h, b) :: map slot2state out) by trivial.
  change (@prod tid (prod heap BarstMap)) with slot in H4.
  rewrite H4 in H0.
  destruct (join_list_uncons H0) as [f [? ?]].
  spec IHout f H5.
  copy H3.
  spec H7 t.
  detach H7.
  2: simpl; auto.
  remember (nth_error thds t).
  icase e.
  icase t0.
  icase th_ctl0.
  clear H7.
  assert (nth_error (tl_states thds) t = Some (empty_stack, th_hp0, th_bs0)).
    unfold tl_states.
    rewrite nth_error_map.
    rewrite <- Heqe.
    trivial.
  destruct (join_assoc _ _ _ _ _ H6 H2) as [f'' [? ?]].
  apply join_com in H8.
  destruct (join_list_join_nth _ _ _ H7 H1 H8) as [stx [tdsts' [? [? ?]]]].
  destruct stx as [[? hx] bx].
  destruct H10 as [[[? ?] ?] ?].
  unfold fst, snd in *.
  subst s.
  clear H10.
  symmetry in Heqe.
  destruct (can_replace _ _ _ _ Heqe (thread th_stk0 hx bx (Running c))) as [thds' ?].
  assert (tdsts' = tl_states thds').
    apply nth_error_eq.
    intro.
    unfold tl_states.
    rewrite nth_error_map.
    case (eq_nat_dec t i); intro.
    subst t.
    apply replace_nth in H11.
    apply replace_nth in H10.
    rewrite H11.
    rewrite H10.
    trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H11 n).
    rewrite (replace_nth2 _ _ _ _ _ _ H10 n).
    unfold tl_states.
    rewrite nth_error_map.
    trivial.
  subst tdsts'.
  spec IHout thds' f'' H12.
  spec IHout tst H9.
  detach IHout.
  (* to IH *)
  destruct IHout as [thds'' ?].
  exists thds''.
  exists hx.
  split.
  apply join_com; trivial.
  exists bx.
  split.
  apply join_com; trivial.
  exists b0.
  exists d.
  exists m.
  exists c.
  split; trivial.
  rewrite H10.
  trivial.
  (* from IH *)
  intros.
  case (eq_nat_dec t i); intro.
  subst i.
  clear -H13 H.
  apply In_nth_error in H13.
  destruct H13 as [j ?].
  rewrite nth_error_map in H0.
  remember (nth_error out j).
  icase e.
  spec H 0 (S j) (t, (h, b)) s.
  detach H; trivial.
  detach H; auto.
  inversion H0.
  rewrite H2 in H.
  detach H; auto.
  disc.
  rewrite (replace_nth2 _ _ _ _ _  _ H10 n).
  apply H3.
  simpl.
  auto.
Qed.

Lemma can_transition_threads: forall bp bn bp' out,
  bp_transition bp bn bp' out ->
  slots_unique_tid_bp bp ->
  forall thds bptl,
  join_list (bp_st bp) (tl_states thds) bptl ->
  (forall dir mv i sx, lookup_slot_bp bp bn dir mv (i, sx) ->
      match nth_error thds i with
        | Some (thread stk hp bs (Waiting bn dir mv c)) => True | _ => False end) ->
  exists thds',
  transition_threads out thds thds'.
Proof.
  intros [bars bst wfbp] ? [bars' bst' wfbp'] [os dir ns ost oslts wfdto].
  intros.
  assert (slts_unique_tids (dbsto_slots (dbsTransOut os dir ns ost oslts wfdto))) by (eapply bpt_uni_out; eauto).
  unfold transition_threads.
  unfold bp_st, dbsto_slots in *.
  destruct H as [? [? ?]].
  unfold lookup_slot_bp in H2.
  unfold bp_bars in *.
  remember (nth_error bars bn).
  icase e.
  destruct d as [bn' dcs wp bd dbswf].
  remember (nth_error bars' bn).
  icase e.
  destruct d as [bn'' dcs' wp' bd' dbswf'].
  destruct wp as [diro wpslts lim wpst wfwp].
  destruct wp' as [diro' wpslts' lim' wpst' wfwp'].
  unfold wp_dir, wp_slots in *.
  destruct H5.
  destruct H5 as [? [dirx [? [? [? [? [? [? [? [? [? [? [? [? [? ?]]]]]]]]]]]]]]].
  subst bn' diro diro' wpslts' lim' wpst' bd' dirx os ns ost.
  spec H2 dir.
  generalize (can_transition_threads' oslts H3); intro.
  spec H5 (st_upd_bm_bn wpst bn'' (barst_builder dcs' pfullshare)) wfdto.
  assert (join bst empty_state bst).
    apply join_com.
    apply join_emptystate.
    copy wfbp.
    destruct wfbp0.
    rewrite <- (join_list_stk _ _ _ H7).
    trivial.
  destruct (join_list_assoc1 H7 H1) as [f [? ?]].
  spec H5 thds f H8.
  copy wfbp.
  destruct wfbp0 as [? _].
  assert (nth_error (dbs_getsts bars) bn = Some wpst).
    unfold dbs_getsts.
    rewrite nth_error_map.
    rewrite <- Heqe.
    trivial.
  assert (join empty_state wpst wpst).
    apply join_emptystate.
    eapply join_list_empty_state; eauto.
  destruct (join_list_split_nth _ _ _ H12 H11 H13) as [f' [obars [? [? ?]]]].
  apply empty_state_identity in H16.
  subst f'.
  apply join_com in H13.
  destruct (join_list_assoc1 H13 H17) as [f'' [? ?]].
  apply join_com in H18.
  destruct (join_assoc _ _ _ _ _ H18 H10) as [f''' [? ?]].
  spec H5 (st_upd_bm_bn f''' bn'' (barst_builder dcs' pfullshare)).
  apply H5.
  (* BR1 *)
  destruct wpst as [[? ?] ?].
  destruct f as [[? ?] ?].
  destruct f''' as [[? ?] ?].
  destruct H21 as [[? ?] ?].
  split.
  split; trivial.
  unfold st_upd_bm_bn, upd_bm, fst, snd in *.
  intro bnx.
  case (eq_nat_dec bnx bn'').
  intro.
  subst bnx.
  clear -H24 H14.
  destruct H14 as [F ?].
  inv_sc H.
  subst s1 s0.
  apply join_com in H2.
  destruct (join_assoc _ _ _ _ _ H2 H24) as [f [? ?]].
  destruct H3 as [? _].
  spec H bn''.
  simpl in H3.
  rewrite H3 in H.
  inversion H.
  constructor.
  destruct y.
  destruct z.
  simpl in H8.
  destruct H8.
  destruct x0.
  destruct x1.
  unfold fst, snd in *.
  change Share.top with (projT1 pfullshare) in H9.
  generalize (nidentity_pair2 n); intro.
  change t with (projT1 (existT (fun t => ~identity t) _ H10)) in H9.
  pfullshare_join.
  intro.
  apply H24.
  (* BR2 *)
  clear -H2 H20.
  intros.
  apply In_nth_error in H.
  destruct H as [j ?].
  rewrite nth_error_map in H.
  spec H2 j i.
  red in H20.
  apply (eq_nth_error j) in H20.
  do 3 rewrite nth_error_map in H20.
  icase (nth_error oslts j).
  destruct s as [? [? ?]].
  inversion H.
  subst t.
  simpl in H20.
  icase (nth_error wpslts j).
  icase o.
  destruct s as [? [? ?]].
  inversion H20.
  subst t.
  spec H2 (h0, b0).
  detach H2; auto.
Qed.

Lemma trans_thds_rev: forall out thds thds',
  transition_threads out thds thds' ->
  forall tid stk hp bs bn dir mv c,
  nth_error thds' tid = Some (thread stk hp bs (Waiting bn dir mv c)) ->
  nth_error thds tid = Some (thread stk hp bs (Waiting bn dir mv c)).
Proof.
  intro.
  destruct out.
  unfold transition_threads.
  simpl.
  clear.
  induction dbsto_slots0.
  intros.
  inversion H.
  subst.
  trivial.
  intros.
  simpl in H.
  destruct a as [? [? ?]].
  remember (nth_error thds t).
  icase e.
  destruct t0.
  destruct H as [hx [? [bx [? ?]]]].
  destruct H2 as [bnx [dirx [mvx [kx [? ?]]]]].
  subst th_ctl0.
  remember (replace thds t (thread th_stk0 hx bx (Running kx))).
  icase o.
  spec IHdbsto_slots0 l thds' H3.
  spec IHdbsto_slots0 tid0 stk hp bs bn.
  spec IHdbsto_slots0 dir mv c.
  apply IHdbsto_slots0 in H0.
  symmetry in Heqo.
  case (eq_nat_dec t tid0); intro.
  subst t.
  apply replace_nth in Heqo.
  rewrite Heqo in H0.
  disc.
  rewrite <- (replace_nth2 _ _ _ _ _ _ Heqo n).
  trivial.
Qed.

Lemma transthds_join: forall out thds thds',
  transition_threads out thds thds' ->
  forall st,
  join_list (dbsto_st out) (tl_states thds) st ->
  join_list empty_state (tl_states thds') st.
Proof.
  intro.
  destruct out.
  unfold transition_threads.
  simpl.
  revert dbsto_st0 dbsto_wf0.
  clear.
  rename dbsto_slots0 into slts.
  induction slts.
  intro st.
  intros.
  inversion dbsto_wf0.
  subst.
  inversion H.
  subst.
  trivial.
  intro st.
  intros.
  destruct a as [? [? ?]].
  simpl in dbsto_wf0.
  apply join_list_uncons in dbsto_wf0.
  destruct dbsto_wf0 as [sltsx [? ?]].
  spec IHslts sltsx H1.
  simpl in H.
  remember (nth_error thds t).
  icase e.
  destruct t0.
  destruct H as [hx [? [bx [? [bnx [drix [mvx [cx [? ?]]]]]]]]].
  remember (replace thds t (thread th_stk0 hx bx (Running cx))).
  icase o.
  spec IHslts l thds' H5.
  apply IHslts.
  clear IHslts.
  assert (nth_error (tl_states thds) t = Some (empty_stack, th_hp0, th_bs0)).
    unfold tl_states.
    rewrite nth_error_map.
    rewrite <- Heqe.
    trivial.
  assert (join empty_state (empty_stack, th_hp0, th_bs0) (empty_stack, th_hp0, th_bs0)).
    apply join_emptystate; auto.
  destruct (join_list_split_nth _ _ _ H6 H0 H7) as [f [thdx [? [? ?]]]].
  destruct (join_assoc _ _ _ _ _ H2 H9) as [f' [? ?]].
  assert (join (empty_stack, h, b) (empty_stack, th_hp0, th_bs0) (empty_stack, hx, bx)).
    repeat (split; trivial).
  apply (join_eq _ _ _ _ H13) in H11.
  subst f'.
  apply join_com in H12.
  destruct (join_list_assoc1 H12 H10) as [f' [? ?]].
  assert (nth_error thdx t = Some empty_state).
    eapply replace_nth; eauto.
  apply join_com in H14.
  destruct (join_list_join_nth _ _ _ H15 H11 H14) as [thstx [lx' [? [? ?]]]].
  replace (tl_states l) with lx'; trivial.
  apply nth_error_eq.
  intro.
  unfold tl_states.
  rewrite nth_error_map.
  symmetry in Heqo.
  case (eq_nat_dec t i); intro.
  subst t.
  apply replace_nth in Heqo.
  rewrite Heqo.
  simpl.
  apply empty_state_identity in H16.
  subst.
  eapply replace_nth; eauto.
  rewrite (replace_nth2 _ _ _ _ _ _ Heqo n).
  rewrite (replace_nth2 _ _ _ _ _ _ H17 n).
  rewrite (replace_nth2 _ _ _ _ _ _ H8 n).
  unfold tl_states.
  rewrite nth_error_map.
  trivial.
Qed.

Definition cs_wellformed (s : sched) (tl : list Thread) (bp : Barpool) (ap : heap) : Prop :=
  (forall bn dir mv tid,
      (exists h, exists b, lookup_slot_bp bp bn dir mv (tid, (h,b))) =
      (exists stk, exists hp, exists bm, exists c, 
        nth_error tl tid = Some (thread stk hp bm (Waiting bn dir mv c)))) /\
  pred_brk (empty_stack, ap, empty_bm) /\ 
    exists bptl, join_list (bp_st bp) (tl_states tl) bptl /\
    exists tw, join (empty_stack, ap, empty_bm) bptl tw /\
      (forall bn ps cst,
        (sepconj (barrier_assert bn ps cst) Tru) tw ->
        option_map dbs_cs (nth_error (bp_bars bp) bn) = Some cst).

Record Cstate : Type := cstate {
  cs_sched : sched;
  cs_thds : list Thread;
  cs_barpool : Barpool;
  cs_allocpool : heap;
  cs_wf : cs_wellformed cs_sched cs_thds cs_barpool cs_allocpool}.

Lemma lookup_move_waiting: forall cs i stk h bs bn dir mv c,
  nth_error (cs_thds cs) i = Some (thread stk h bs (Waiting bn dir mv c)) ->
  exists pre, exists post,
    lookup_move_bp (cs_barpool cs) bn dir mv pre post.
Proof.
  intros.
  destruct cs.
  simpl in *.
  destruct cs_wf0 as [? _].
  spec H0 bn dir mv i.
  assert (exists stk, exists hp, exists bm, exists c, 
                 nth_error cs_thds0 i = Some (thread stk hp bm (Waiting bn dir mv c))).
    exists stk.
    exists h.
    exists bs.
    exists c.
    trivial.
  rewrite <- H0 in H1.
  destruct H1 as [hx [bx ?]].
  clear -H1.
  unfold lookup_move_bp.
  red in H1.
  destruct cs_barpool0.
  simpl in *.
  icase (nth_error bp_bars0 bn).
  destruct d as [bn' cs wp bd wfdbs].
  destruct H1.
  destruct wp as [diro slts lim st wfwp].
  simpl in *.
  subst diro.
  destruct wfdbs as [? [? ?]].
  spec H2 mv.
  rewrite H0 in H2.
  destruct H2 as [pre [post [? ?]]].
  exists pre.
  exists post.
  trivial.
Qed.

Lemma cs_slots_unique: forall cs,
  slots_unique_tid_bp (cs_barpool cs).
Proof.
  intros [sch thds bp ap wfcs]; red; intros.
  simpl in *.
  destruct wfcs as [? _].
  copy H2.
  destruct slt as [? [? ?]].
  destruct slt' as [? [? ?]].
  simpl in *.
  subst t0.
  spec H2 bn dir mv t.
  spec H3 bn' dir' mv' t.
  assert (exists h, exists b, lookup_slot_bp bp bn dir mv (t, (h, b))).
    exists h.
    exists b.
    trivial.
  assert (exists h, exists b, lookup_slot_bp bp bn' dir' mv' (t, (h, b))).
    exists h0.
    exists b0.
    trivial.
  rewrite H3 in H4.
  rewrite H2 in H1.
  destruct H1 as [? [? [? [? ?]]]].
  destruct H4 as [? [? [? [? ?]]]].
  rewrite H1 in H4.
  inversion H4.
  subst.
  repeat (split; trivial).
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H) in H0.
  trivial.
Qed.

Lemma cs_running_nin_bp : forall cs t stk hp bs c,
  nth_error (cs_thds cs) t = Some (thread stk hp bs (Running c)) ->
  ~tid_in_bp (cs_barpool cs) t.
Proof.
  repeat intro.
  destruct cs.
  simpl in *.
  destruct H0 as [bn [dir [mv [slt [? ?]]]]].
  destruct cs_wf0 as [? _].
  spec H2 bn dir mv t.
  assert (exists h, exists b, lookup_slot_bp cs_barpool0 bn dir mv (t, (h, b))).
    destruct slt as [? [? ?]].
    exists h.
    exists b.
    simpl in H1.
    subst t0.
    trivial.
  change nat with tid in H3.
  rewrite H2 in H3.
  destruct H3 as [? [? [? [? ?]]]].
  rewrite H in H3.
  disc.
Qed.

Lemma lupst_joins_thd: forall cs bn st,
  lookup_state_bp (cs_barpool cs) bn st ->
  forall i stk hp bs cctl,
  nth_error (cs_thds cs) i = Some (thread stk hp bs cctl) ->
  joins st (empty_stack, hp, bs).
Proof.
  intros.
  destruct cs.
  simpl in H, H0.
  destruct cs_wf0 as [_ [_ ?]].
  destruct H1 as [bptl [? _]].
  destruct cs_barpool0 as [bars stx wfbp].
  unfold lookup_state_bp, bp_bars, bp_st in *.
  remember (nth_error bars bn).
  icase e.
  destruct wfbp.
  destruct d.
  destruct dbs_wp0.
  subst wp_st0.
  (* should lemma-ize *)
  assert (nth_error (dbs_getsts bars) bn = Some st).
    unfold dbs_getsts.
    rewrite nth_error_map.
    rewrite <- Heqe.
    unfold compose.
    trivial.
  assert (join empty_state st st).
    apply join_emptystate.
    eapply join_list_empty_state; eauto.
  destruct (join_list_split_nth _ _ _ H H2 H4) as [? [bars' [? [? ?]]]].
  apply empty_state_identity in H6.
  subst x.
  apply join_com in H4.
  destruct (join_list_assoc1 H4 H7) as [bast' [? ?]].
  clear -H8 H0 H1 H4.
  assert (nth_error (tl_states cs_thds0) i = Some (empty_stack, hp, bs)).
    unfold tl_states.
    rewrite nth_error_map.
    rewrite H0.
    trivial.
  assert (join empty_state (empty_stack, hp, bs) (empty_stack, hp, bs)).
    apply join_emptystate.
    trivial.
  destruct (join_list_split_nth _ _ _ H H1 H2) as [? [thds' [? [? ?]]]].
  apply join_com in H8.
  destruct (join_assoc _ _ _ _ _ H8 H5) as [f [? ?]].
  exists f.
  trivial.
Qed.

(* An ugly but needed variant *)
Lemma lupst_joins_thd2: forall cs bn st,
  lookup_state_bp (cs_barpool cs) bn st ->
  forall i stk hp bs cctl i' stk' hp' bs' cctl',
  i <> i' ->
  nth_error (cs_thds cs) i = Some (thread stk hp bs cctl) ->
  nth_error (cs_thds cs) i' = Some (thread stk' hp' bs' cctl') ->
  exists thsts,
  join (empty_stack, hp, bs) (empty_stack, hp', bs') thsts /\
  joins st thsts.
Proof.
  intros.
  destruct cs.
  simpl in *.
  destruct cs_wf0 as [_ [_ [btst [? _]]]].
  apply tl_states_lookup in H1.
  apply tl_states_lookup in H2.
  assert (join empty_state (empty_stack, hp, bs) (empty_stack, hp, bs)) by (apply join_emptystate; auto).
  assert (join empty_state (empty_stack, hp', bs') (empty_stack, hp', bs')) by (apply join_emptystate; auto).
  destruct (join_list_split_nth _ _ _ H1 H3 H4) as [x [sts' [? [? ?]]]].
  generalize (replace_nth2 _ _ _ _ _ _ H6 H0); intro.
  rewrite H2 in H9.
  destruct (join_list_split_nth _ _ _ H9 H8 H5) as [x' [sts'' [? [? ?]]]].
  destruct (join_assoc _ _ _ _ _ H7 H11) as [f [? ?]].
  exists f.
  split; trivial.
  clear -H14 H.
  destruct cs_barpool0 as [bars bpst [wfbp ?]].
  red in H.
  unfold bp_st, bp_bars in *.
  clear e.
  remember (nth_error bars bn).
  icase e.
  destruct d as [bnx cs [diro slts lim bst] bd wfdbs].
  subst st.
  unfold dbs_getsts in wfbp.
  assert (nth_error (map (wp_st oo dbs_wp) bars) bn = Some bst).
    rewrite nth_error_map.
    rewrite <- Heqe.
    reflexivity.
  assert (join empty_state bst bst).
    apply join_emptystate.
    eapply join_list_empty_state; eauto.
  destruct (join_list_split_nth _ _ _ H wfbp H0) as [x'' [bsts [? [? ?]]]].
  apply empty_state_identity in H2.
  subst x''.
  apply join_com in H0.
  destruct (join_list_assoc1 H0 H3) as [f' [? ?]].
  apply join_com in H4.
  destruct (join_assoc _ _ _ _ _ H4 H14) as [f'' [? ?]].
  exists f''.
  trivial.
Qed.
