(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Tactic Notation "inv_sc" hyp(H) := 
  (destruct H as [[[? ?] ?] [[[? ?] ?] [[[[? ?] ?] ?] [? ?]]]]; unfold fst, snd in *).

(* SYNTAX *)

Definition addr : Type := nat.
Definition data : Type := nat.
Definition bar_num : Type := nat.
Definition var : Type := nat.

Inductive value : Type :=
 | V_data : data -> value
 | V_addr : addr -> value
 |truE : value
 |falsE : value.

Inductive expr : Type :=
 | expConst : value -> expr
 | expVar : var -> expr.

Inductive cmd : Type :=
  | Barrier : bar_num -> cmd
  | Assign : var -> expr -> cmd
  | Load(*from. means read*) : var -> expr -> cmd
  | Store(* to. means write*) : expr -> expr -> cmd 
  | Seq : cmd -> cmd -> cmd
  | If : expr -> cmd -> cmd -> cmd
  | While : expr -> cmd -> cmd
  | New : var -> expr -> cmd
  | Free : expr -> cmd
  | Unit : cmd.

(* BASE SEMANTICS & SEPARATION ALGEBRAS *)

(* We could do this in a more semantic style, but why bother? *)
Fixpoint modifies (c : cmd) (x : var) : Prop :=
  match c with
   | Assign x' _ => x = x'
   | Load x' _ => x = x'
   | New x' _ => x = x'
   | Seq c1 c2 => (modifies c1 x) \/ (modifies c2 x)
   | If _ c1 c2 => (modifies c1 x) \/ (modifies c2 x)
   | While _ c => modifies c x
   | _ => False
  end.

Fixpoint modifies_bn (c : cmd) (bn : bar_num) : Prop :=
  match c with
   | Seq c1 c2 => (modifies_bn c1 bn) \/ (modifies_bn c2 bn)
   | If _ c1 c2 => (modifies_bn c1 bn) \/ (modifies_bn c2 bn)
   | While _ c => modifies_bn c bn
   | Barrier bn' => bn = bn'
   | _ => False
  end.

Definition sa_value : sepalg value := 
  sa_equiv value.

(* Stacks *)
Definition stack : Type := var -> value.

Definition upd_stack (stk : stack) (snum : var) (newval : value) : stack :=
  fun num => if eq_nat_dec snum num then newval else stk num.

Lemma upd_stk_eq: forall stk x v,
  upd_stack stk x v x = v.
Proof.
  unfold upd_stack.
  intros.
  case (eq_nat_dec x x); auto.
  intros.
  contradiction n.
  trivial.
Qed.

Lemma upd_stk_neq: forall stk x v x',
  x <> x' ->
  upd_stack stk x v x' = stk x'.
Proof.
  unfold upd_stack.
  intros.
  case (eq_nat_dec x x'); intro; trivial; contr.
Qed.

Definition empty_stack : stack := fun _ => falsE.

Definition sa_stack : sepalg stack :=
  sa_equiv stack.
Existing Instance sa_stack.

Fixpoint evalExpr (rho : stack) (e : expr) : value :=
  match e with
   | expConst v => v
   | expVar x => rho x
  end.

(* Heaps *)

Definition brk : Type := option nat.

Definition sa_brk : sepalg brk :=
  sa_option _.
Existing Instance sa_brk.

Definition share : Type := Share.t.
Definition fullshare : share := Share.top.
Definition emptyshare : share := Share.bot.

Definition res : Type := 
  option (pjoinable (sa_prod sa_value Share.sa)).

Definition res_builder (v : value) (psh : pshare) : res := 
  match psh with 
   (existT sh p_sh_ne) => 
     Some (existT _ (v, sh) (n2_identity_prod p_sh_ne)) 
  end.

Definition sa_res : sepalg res :=
  sa_lift (sa_prod sa_value Share.sa).
Existing Instance sa_res.

Definition heap : Type := (brk * (addr -> res))%type.
Definition upd_heap (hp : heap) (adr : addr) (newval : res): heap :=
  (fst hp, fun a => if eq_nat_dec a adr then newval else (snd hp) a).

Definition brk_heap (hp : heap) : brk := 
  fst hp.

Definition dref_heap (hp : heap) (adr :addr) : res :=
  (snd hp) adr.

Definition new_heap (hp : heap) (r : res) : option (heap * addr) :=
  match hp with
    | (Some brk, hf) =>
        Some ((Some (brk + 1), 
                      fun a => if eq_nat_dec a brk then r else hf a), brk)
    | (None, _) => None
  end.

Definition empty_heap : heap := 
  (None, fun addr => None).

Definition sa_heap : sepalg heap :=
  sa_prod sa_brk (sa_fun _ sa_res).
Existing Instance sa_heap.

Lemma empty_heap_identity: identity empty_heap.
Proof.
  rewrite identity_unit_equiv.
  unfold empty_heap.
  split; simpl.
  unfold ojoin.
  trivial.
  intro.
  constructor.
Qed.

Lemma join_emptyheap: forall h,
  join empty_heap h h.
Proof.
  intro.
  unfold empty_heap.
  split; simpl.
  unfold ojoin.
  destruct (fst h); trivial.
  intro.
  constructor.
Qed.

Lemma identity_emptyheap: forall h,
  identity h ->
  h = empty_heap.
Proof.
  intros.
  unfold empty_heap.
  destruct h.
  rewrite identity_unit_equiv in H.
  destruct H.
  icase b.
  clear H.
  apply injective_projections; trivial.
  unfold snd in *.
  extensionality a.
  spec H0 a.
  icase (r a); trivial.
  destruct p.
  inversion H0.
  simpl in H3.
  contradiction n.
  rewrite identity_unit_equiv.
  apply H3.
Qed.

(* Barrier state *)
Definition stateid : Type := nat.
Definition sa_stateid : sepalg stateid :=
  sa_equiv stateid.

Definition barst : Type := 
  option (pjoinable (sa_prod sa_stateid Share.sa)).

Definition barst_builder (sid : stateid) (psh : pshare) : barst := 
  match psh with 
   (existT sh p_sh_ne) => 
     Some (existT _ (sid, sh) (n2_identity_prod p_sh_ne)) 
  end.

Definition sa_barst : sepalg barst :=
  sa_lift (sa_prod sa_stateid Share.sa).
Existing Instance sa_barst.

Definition BarstMap : Type := bar_num -> barst.
Definition sa_bsm : sepalg BarstMap :=
  sa_fun _ sa_barst.
Existing Instance sa_bsm.

Definition upd_bm (bm : BarstMap) (bn : bar_num) (newbst : barst): BarstMap :=
  fun a => if eq_nat_dec a bn then newbst else bm a.

Definition empty_bm : BarstMap := 
  fun bn => None.

Lemma empty_bm_identity: identity empty_bm.
Proof.
  rewrite identity_unit_equiv.
  unfold empty_bm.
  intro.
  constructor.
Qed.

Lemma join_emptybm: forall h,
  join empty_bm h h.
Proof.
  intro.
  unfold empty_bm.
  intro.
  constructor.
Qed.

Lemma identity_emptybm: forall h,
  identity h ->
  h = empty_bm.
Proof.
  intros.
  unfold empty_bm.
  extensionality a.
  rewrite identity_unit_equiv in H.
  spec H a.
  icase (h a); trivial.
  destruct p.
  inversion H.
  simpl in H3.
  contradiction n.
  rewrite identity_unit_equiv.
  apply H3.
Qed.

(* States *)

Definition state : Type := (stack * heap * BarstMap)%type.

Definition st_stk (st :state) : stack := fst (fst st).
Definition st_hp (st : state) : heap := snd (fst st).
Definition st_hf (st : state) : addr -> res := snd (st_hp st).
Definition st_brk (st : state) : brk := fst (st_hp st).
Definition st_bsm (st : state) : BarstMap := snd st.

Definition st_upd_stk (st :state) (stk : stack) : state := 
  (stk, snd (fst st), snd st).

Definition st_upd_stk_var (st :state) (x : var) (v : value) : state := 
  st_upd_stk st (upd_stack (st_stk st) x v).

Definition st_upd_bm_bn (st : state) (bn : bar_num) (bs : barst) : state :=
  (fst (fst st), snd (fst st), upd_bm (snd st) bn bs).

Definition empty_state : state := (empty_stack, empty_heap, empty_bm).

Definition sa_state : sepalg state :=
  sa_prod (sa_prod sa_stack sa_heap) sa_bsm.
Existing Instance sa_state.

Lemma empty_state_identity: identity empty_state.
Proof.
  rewrite identity_unit_equiv.
  unfold empty_state.
  split.
  split.
  split; trivial.
  apply join_emptyheap.
  apply join_emptybm.
Qed.

Lemma join_emptystate: forall s,
  st_stk s = empty_stack ->
  join empty_state s s.
Proof.
  intros.
  unfold empty_state.
  split.
  split.
  simpl.
  destruct s as [[? ?] ?].
  compute in H.
  subst s.
  split; trivial.
  apply join_emptyheap.
  apply join_emptybm.
Qed.

Lemma identity_emptystate: forall s,
  identity s ->
  st_stk s = empty_stack ->
  s = empty_state.
Proof.
  intros.
  destruct s as [[? ?] ?].
  rewrite identity_unit_equiv in H.
  destruct H as [[? ?] ?].
  unfold st_stk, fst, snd in *.
  unfold empty_state.
  apply injective_projections; simpl.
  apply injective_projections; simpl.
  trivial.
  apply identity_emptyheap.
  rewrite identity_unit_equiv.
  trivial.
  apply identity_emptybm.
  rewrite identity_unit_equiv.
  trivial.
Qed.

Lemma join_stk_eq: forall st1 st2 st,
join st1 st2 st -> st_stk st1 = st_stk st.
Proof.
  intros.
   destruct H.
   simpl in *.
   destruct H.
   simpl in *.
   destruct H.
   unfold st_stk.
   congruence.
Qed.

(* ASSERTIONS *)

Definition assert : Type := state -> Prop.

Definition derives (P Q : assert) : Prop :=
  forall st, P st -> Q st.

(* Pure Logic *)

Definition Fls : assert :=
  fun st => False.

Definition Tru : assert :=
  fun st => True.

Definition and (P Q : assert) : assert :=
  fun st => P st /\ Q st .

Definition or (P Q : assert) : assert := 
  fun st => P st \/ Q st.

Definition impl (P Q : assert) :=
  fun st => P st -> Q st.

Definition not (P : assert) :=
  impl P Fls.

Definition exists_ (A : Type) (P : A -> assert) : assert :=
  fun st => exists a, P a st.
Implicit Arguments exists_.
Notation "'Ex_' x ':' T ',' P " := (exists_ (fun x : T => P)) (at level 65, x at level 99).

Definition forall_ (A: Type) (P: A -> assert) : assert :=
  fun st => forall a : A, P a st.
Implicit Arguments forall_.
Notation "'All_' x ':' T ',' P " := (forall_ (fun x : T => P)) (at level 65, x at level 99).

Definition exact (st : state) : assert :=
  fun st' => st = st'.

(* Pure Separation Logic *)

Definition emp : assert :=
  identity.

Definition sepconj (P Q : assert) : assert :=
  fun st => exists st1, exists st2,
    join st1 st2 st /\ P st1 /\ Q st2.

Lemma sep_eq: forall P Q R,
  Q = R ->
  sepconj P Q = sepconj P R.
Proof.
  intros.
  rewrite H.
  trivial.
Qed.

Notation "P ** Q" := (sepconj P Q)(at level 200).

Lemma emp_simpl : forall P , sepconj P emp = P.
Proof.
  intro.
  extensionality.
  apply prop_ext.
  split; intro.
  do 3 destruct H.
  destruct H0.
  spec H1 x0 x.
  apply join_com in H.
  spec H1 H; subst;trivial.
  exists x.
  exists (st_stk x , empty_heap, empty_bm).
  split.
  destruct x as [[? ?] ?].
  split.
  unfold st_stk , fst.
  split.
  unfold fst.
  split ; trivial.
  unfold snd.
  apply join_com.
  apply join_emptyheap.
  unfold snd.
  apply join_com.
  apply join_emptybm.
  split; trivial.
  repeat intro.
  destruct H0.
  destruct a.
  destruct b.     
  unfold fst, snd in *.
  apply injective_projections.
  unfold fst.
  destruct p.
  destruct p0.
  apply injective_projections.
  unfold fst.
  destruct H0.
  destruct H0;trivial.
  destruct H0.
  unfold snd in *.
  apply (empty_heap_identity _ _ H2).
  apply (empty_bm_identity _ _ H1).
Qed.

Lemma sc_comm :  forall (P Q : assert),
  sepconj P Q = sepconj Q P.
Proof.
  intros.
  extensionality Drunk.
  apply prop_ext; split; intros.
  destruct H as [beer ?].
  destruct H as [vodka ?].
  destruct H.
  unfold sepconj.
  exists vodka.
  exists beer.
  split; try tauto.
  apply join_com.
  exact H.
  
  destruct H as [vodka ?].
  destruct H as [beer ?].
  unfold sepconj.
  destruct H.
  exists beer.
  exists vodka.
  destruct H0.
  split; auto.
  apply join_com.
  exact H.
Qed.
  
Lemma sc_assoc: forall (P Q R : assert),
  sepconj (sepconj P Q) R = 
  sepconj P (sepconj Q R).

Proof.
  intros.
  extensionality st.
  apply prop_ext; split; intros.

  unfold sepconj in H.
  destruct H as [Hilt ?].
  destruct H as [Blade ?].
  destruct H.
  destruct H0.
  destruct H0 as [Pommel ?].
  destruct H0 as [Cross ?].
  destruct H0.
  
  unfold sepconj.
  generalize (join_assoc Pommel Cross Blade Hilt st H0 H).
  intros.
  destruct X.
  destruct a.
  exists Pommel.
  exists x.
  split; auto.
  split; try tauto.
  exists Cross.
  exists Blade.
  split; tauto.

  unfold sepconj in H.
  destruct H as [Blade ?].
  destruct H as [Hilt ?].
  destruct H.
  destruct H0.
  destruct H1 as [Pommel ?].
  destruct H1 as [Cross ?].
  destruct H1.
  
  apply join_com in H.
  apply join_com in H1.
  generalize (join_assoc _ _ _ _ _ H1 H).
  intros.
  destruct X.
  destruct a.
  apply join_com in H3.
  apply join_com in H4.
  unfold sepconj.
  exists x.
  exists Cross.
  split; auto.
  split; try tauto.
  exists Blade.
  exists Pommel.
  split; tauto.
Qed.

(* New list stuff -- add to MSL *)
Fixpoint sepconjList (L : list assert) : assert :=
  match L with
    | nil => emp
    | P :: Lx => sepconj P (sepconjList Lx)
  end.

Fixpoint assert_list_state_list (alist : list assert) (slist : list state) : Prop :=
  match (alist, slist) with
   | (nil, nil) => True
   | (h :: hl', nil) => False
   | (nil, a :: alist') => False
   | (a :: alist', s :: slist') => a s /\ assert_list_state_list alist' slist'
  end.  

Lemma sepconjList_joinlist: forall L stk h bs,
  (sepconjList L) (stk, h, bs) ->
  exists Lst,
    join_list (stk, empty_heap, empty_bm) Lst (stk, h, bs) /\
    assert_list_state_list L Lst.
Proof.
  induction L; intros.
  simpl in H.
  exists nil.
  simpl.
  split; trivial.
  unfold emp in H.
  rewrite identity_unit_equiv in H.
  destruct H as [[? ?] ?].
  unfold fst, snd in *.
  change (join h h h) with (unit_for h h) in H0.
  change (join bs bs bs) with (unit_for bs bs) in H1.
  rewrite <- identity_unit_equiv in H0.
  rewrite <- identity_unit_equiv in H1.
  replace h with empty_heap.
  replace bs with empty_bm.
  constructor.

  rewrite identity_emptybm; trivial.
  rewrite identity_emptyheap; trivial.

  simpl in H.
  inv_sc H.
  subst.
  spec IHL stk h1 b0 H4.
  destruct IHL as [L' [? ?]].
  exists ((stk, h0, b) :: L').
  split.
  2: split; trivial.
  clear H0 H4.
  eapply join_list_cons; eauto.
  apply join_com.
  repeat (split; trivial).
Qed.

Lemma assertlist_statelist_sepconjlist: forall Ps Ss,
  assert_list_state_list Ps Ss ->
  forall st tst, identity st ->
    join_list st Ss tst ->
      sepconjList Ps tst.
Proof.
  induction Ps; intros.
  icase Ss.
  inversion H1.
  subst.
  trivial.
  icase Ss.
  simpl in H.
  simpl.
  destruct (join_list_uncons H1) as [tst' [? ?]].
  exists s.
  exists tst'.
  apply join_com in H3.
  split; trivial.
  destruct H.
  split; trivial.
  apply IHPs with Ss st; trivial.
Qed.

Lemma assert_list_assert_holds: forall AL SL,
  assert_list_state_list AL SL ->
  forall i A S,
  nth_error AL i = Some A ->
  nth_error SL i = Some S ->
  A S.
Proof.
  induction AL; intros.
  icase i.
  destruct i.
  icase SL.
  inversion H0.
  inversion H1.
  subst A S.
  destruct H.
  trivial.
  icase SL.
  simpl in *.
  destruct H.
  spec IHAL SL H2 i A S.
  auto.
Qed.

Definition sepimpl (P Q : assert) : assert :=
  fun st => forall st' st'': state, join st st' st'' -> P st' -> Q st''.

Definition precise (P : assert) : Prop :=
  forall st st' st'',
    join_sub st' st -> join_sub st'' st ->
      P st' -> P st'' ->
        st' = st''.

Lemma precise_sep: forall P Q, precise P -> precise Q -> precise (sepconj P Q).
Proof.
  repeat intro.
  unfold precise in H.
  do 3 destruct H3. destruct H5.
  do 3 destruct H4. destruct H7.
  destruct H1.
  destruct (join_assoc _ _ _ _ _ H3 H1) as [? [? ?]].
  destruct H2.
  destruct (join_assoc _ _ _ _ _ H4 H2) as [? [? ?]].
  spec H st x x1.
  spec H0 st x0 x2.
  detach H. detach H0.
  spec H0 H6.  spec H0 H8.
  detach H. detach H.
  spec H H5.  spec H H7.
  subst.
  eapply join_eq; eauto.
  
  
  exists x6. trivial.
  exists x4. trivial.
  apply join_com in H12.
  destruct (join_assoc _ _ _ _ _ H11 H12) as [? [? ?]].
  exists x7. trivial.
  apply join_com in H10.
  destruct (join_assoc _ _ _ _ _ H9 H10) as [? [? ?]].
  exists x7. trivial.
Qed.

(* Language Based *)

Definition everywhere (P : assert) : assert :=
  fun st => forall stk, P (st_upd_stk st stk).

Definition ubiquitous (P : assert) : Prop :=
  everywhere P = P.

Lemma emp_ubiquitous : ubiquitous emp.
Proof.
  unfold ubiquitous, everywhere.
  extensionality st.
  apply prop_ext; split; intro; auto.
  destruct st as [[? ?] ?].
  spec H s.
  auto.
  intro.
  destruct st.
  destruct p.
  unfold emp in *.
  unfold st_upd_stk, fst, snd.
  rewrite identity_unit_equiv.
  rewrite identity_unit_equiv in H.
  destruct H as [[_ ?] ?].
  repeat (split; trivial).
Qed.

Lemma ubiquitous_sep: forall P Q, ubiquitous P -> ubiquitous Q -> ubiquitous (sepconj P Q).
   Proof.
     repeat intro.
     unfold ubiquitous, everywhere in *.
     rewrite <- H.
     rewrite <- H0.

     extensionality.

     apply prop_ext.
     split.
     intro.
     spec H1 (st_stk x).
     destruct x as [[? ?] ?].
     unfold st_stk, st_upd_stk,fst,snd in H1;trivial.
     repeat intro.
     do 3 destruct H1. destruct H2.
     exists (st_upd_stk x0 stk).
     exists (st_upd_stk x1 stk).
     destruct H1.
     destruct H1.
     destruct x0 as [[? ?] ?].
     destruct x1 as [[? ?] ?].
     destruct x  as [[? ?] ?].
     unfold fst, snd in *.
     destruct H1. subst s0 s1.
     unfold fst,st_upd_stk,fst,snd.
     split. split;trivial. unfold fst,snd in *.
     split;trivial.
     split;trivial.
     split. intro.
     spec H2 stk0. trivial.
     intro.
     spec H3 stk0. trivial.     
   Qed.

Definition eval_assert (e : expr) (v : value) : assert :=
  fun st => 
    (evalExpr (st_stk st) e = v) /\ 
    (st_hp st = empty_heap) /\
    (st_bsm st = empty_bm).

Definition barrier_assert (bn : bar_num) (ps : pshare) (bs : stateid) : assert :=
  fun st =>
    st_bsm st bn = barst_builder bs ps /\
    (forall bn', bn' <> bn -> st_bsm st bn' = None) /\
    st_hp st = empty_heap.

Lemma ubiquitous_barrier_assert: forall bn ps cst,
  ubiquitous (barrier_assert bn ps cst).
Proof.
  intros.
  unfold ubiquitous.
  extensionality st.
  apply prop_ext; split; intro.
  spec H (st_stk st).
  apply H.
  intro.
  apply H.
Qed.

Lemma precise_barrier : forall bn sh bst, precise (barrier_assert bn sh bst).
Proof.
  repeat intro.
  unfold barrier_assert in *.
  destruct st' as [[? [? ?]] ?].
  destruct st'' as [[? [? ?]] ?].
  unfold st_brk,st_hf, st_hp,st_bsm, fst,snd in *.
  destruct H2. destruct H3. destruct H4.
  destruct H1. destruct H4. destruct H5.
  do 4 destruct H. 
  do 4 destruct H0.
  destruct st as [[? [? ?]] ?].
  destruct x as [[? [? ?]] ?].
  destruct x0 as [[? [? ?]] ?].
  unfold fst,snd in *;subst.
  apply injective_projections; trivial. unfold snd.
  extensionality.
  LEM (x=bn);subst.
  rewrite H1. rewrite H2.  trivial.
  spec H4 x. spec H3 x.
  spec H3 H. spec H4 H.
  rewrite H3. rewrite H4. trivial.
Qed.

Lemma barrier_assert_split: forall bn (p1 p2 p3 : pshare) cs,
  pjoin p1 p2 p3 ->
  (sepconj (barrier_assert bn p1 cs) (barrier_assert bn p2 cs)) = barrier_assert bn p3 cs.
Proof.
  intros.
  extensionality st.
  destruct st as [[? ?] ?].
  unfold barrier_assert.
  apply prop_ext; split; intro.
  inv_sc H0.
  subst s0 s1.
  destruct H5 as [? [? ?]].
  destruct H4 as [? [? ?]].
  unfold st_bsm, st_hp, fst, snd in *.
  subst h0 h1.
  split.
  spec H3 bn.
  rewrite H0, H4 in H3.
  destruct p1.
  destruct p2.
  destruct p3.
  unfold barst_builder in *.
  inversion H3; subst.
  simpl in H9.
  destruct z.
  destruct x2.
  destruct H9 as [[? ?] ?].
  unfold fst, snd in *.
  simpl in H.
  generalize (join_eq _ _ _ _ H H9); intro.
  subst x1.
  subst s0.
  replace n2 with (@n2_identity_prod _ _ sa_stateid Share.sa cs t n1) by apply proof_irr.
  trivial.
  split.
  intros.
  spec H3 bn'.
  spec H1 bn' H5.
  spec H6 bn' H5.
  rewrite H6, H1 in H3.
  inversion H3; auto.
  generalize (empty_heap_identity _ _ H2).
  auto.
  
  destruct H0 as [? [? ?]].
  unfold st_hp, st_bsm, fst, snd in *.
  subst h.
  exists (s, empty_heap, upd_bm b bn (barst_builder cs p1)).
  exists (s, empty_heap, upd_bm b bn (barst_builder cs p2)).
  split.
  split.
  split.
  split; trivial.
  apply join_emptyheap.
  intro.
  unfold snd, upd_bm.
  case (eq_nat_dec x bn); intro.
  subst x.
  rewrite H0.
  unfold barst_builder.
  destruct p1.
  destruct p2.
  destruct p3.
  constructor.
  split; trivial.
  split; trivial.
  spec H1 x n.
  rewrite H1.
  constructor.
  unfold upd_bm.
  case (eq_nat_dec bn bn); intro.
  2: contradiction n; trivial.
  split; split; trivial; split; trivial; intros; case (eq_nat_dec bn' bn); intro; contr; spec H1 bn' n; trivial.
Qed.

Definition isTru_assert  ex : assert :=
  eval_assert ex truE.

Definition points2 (e : expr) (sh : pshare) (e' : expr) : assert :=
    fun st => 
      match evalExpr (st_stk st) e with
        | V_data _ => False
        | V_addr a => 
          st_brk st = None /\
          (st_hf st a) = res_builder (evalExpr (st_stk st) e') sh /\
          (forall x , x <>a  -> st_hf st x = None) /\
          st_bsm st = empty_bm
        |truE  => False
        |falsE => False
    end.

Lemma points2_split: forall e (p1 p2 p3 : pshare) e',
  pjoin p1 p2 p3 ->
  (sepconj (points2 e p1 e') (points2 e p2 e')) = points2 e p3 e'.
Proof.
  intros.
  extensionality st.
  destruct st as [[? ?] ?].
  unfold points2.
  apply prop_ext; split; intro.
  inv_sc H0.
  subst s0 s1.
  unfold st_stk, fst in *.
  remember (evalExpr s e).
  icase v.
  destruct H5 as [? [? [? ?]]].
  destruct H4 as [? [? [? ?]]].
  unfold st_bsm, st_hf, st_brk, st_hp, snd, fst in *.
  subst b0 b1.
  destruct h.
  destruct h1.
  destruct h0.
  subst b1 b2.
  destruct H2.
  unfold fst, snd in *.
  icase b0.
  split; trivial.
  apply empty_bm_identity in H3.
  subst b.

  split.
  spec H2 a.
  rewrite H1, H7 in H2.
  destruct p1.
  destruct p2.
  destruct p3.
  unfold res_builder in *.
  inversion H2; subst.
  simpl in H9.
  destruct z.
  destruct x2.
  destruct H9 as [[? ?] ?].
  unfold fst, snd in *.
  simpl in H.
  generalize (join_eq _ _ _ _ H H9); intro.
  subst x1 v.
  replace n2 with (@n2_identity_prod _ _ sa_value Share.sa (evalExpr s e') t n1) by apply proof_irr.
  trivial.
  split; trivial.
  intros.
  spec H2 x.
  spec H5 x H3.
  spec H8 x H3.
  rewrite H5, H8 in H2.
  inversion H2; auto.
  
  unfold st_stk, fst in *.
  remember (evalExpr s e).
  icase v.
  destruct H0 as [? [? [? ?]]].
  unfold st_hf, st_brk, st_hp, st_bsm, fst, snd in *.
  subst b.
  exists (s, upd_heap h a (res_builder (evalExpr s e') p1), empty_bm).
  exists (s, upd_heap h a (res_builder (evalExpr s e') p2), empty_bm).
  rewrite <- Heqv.
  split.
  split.
  split.
  split; trivial.
  destruct h.
  subst b.
  split.
  simpl.
  red.
  trivial.
  unfold snd, fst, upd_heap.
  intro.
  unfold snd.
  case (eq_nat_dec x a); intro.
  subst x.
  rewrite H1.
  unfold res_builder.
  destruct p1.
  destruct p2.
  destruct p3.
  constructor.
  split; trivial.
  split; trivial.
  spec H2 x n.
  rewrite H2.
  constructor.
  apply join_emptybm.
  unfold upd_heap.
  case (eq_nat_dec a a); intro.
  2: contradiction n; trivial.
  split; split; trivial; split; trivial; split; trivial; intros; case (eq_nat_dec x a); intro; contr; spec H2 x n; trivial.
Qed.

Lemma precise_points2 : forall e1 e2 sh, (precise (points2 e1 sh e2)).
Proof.
  repeat intro.
  unfold points2 in *.
  remember (evalExpr (st_stk st') e1).
  icase v.
  remember (evalExpr (st_stk st'') e1).
  icase v.
  destruct st as [[? [? ?]] ?].
  destruct st' as [[? [? ?]] ?].
  destruct st'' as [[? [? ?]] ?].
  destruct H1 as [? [? [? ?]]].
  destruct H2 as [? [? [? ?]]].
  unfold st_stk, st_brk,st_hf, st_hp,st_bsm, fst,snd in *.
  subst.
  destruct H as [[[sx [bkx hfx]] bx] [[[? ?] [? ?]] ?]].
  destruct H0 as [[[sx' [bkx' hfx']] bx'] [[[? ?] [? ?]] ?]].
  unfold fst, snd in *.
  subst s1 s0 sx sx'.
  replace r0 with r1; trivial.
  extensionality adr.
  rewrite <- Heqv0 in Heqv.
  inversion Heqv.
  subst a0.
  case (eq_nat_dec adr a); intro.
  subst a.
  congruence.
  spec H4 adr n.
  spec H7 adr n.
  congruence.
Qed.

Definition pred_brk : assert :=
  fun st =>
    st_bsm st = empty_bm /\
    exists m, st_brk st = Some m /\
      forall n,
        ((n >= m) -> ((st_hf st) n) = res_builder falsE pfullshare) /\
        ((n < m) ->
        ((st_hf st) n) = None).

Lemma pred_brk_precise : precise pred_brk.
Proof.
  intros [[? [? ?]] ?] [[? [? ?]] ?] [[? [? ?]] ?].
  intros [[[? [? ?]] ?] [[[? ?] [? ?]] ?]].
  intros [[[? [? ?]] ?] [[[? ?] [? ?]] ?]].
  intros [? [? [? ?]]] [? [? [? ?]]].
  unfold st_brk, st_bsm, st_hf, st_hp, fst, snd in *.
  subst.
  apply injective_projections; trivial.
  apply injective_projections; trivial.
  simpl.
  icase b7.
  icase b5.
  icase b.
  compute in H1, H6.
  subst.
  apply injective_projections; trivial.
  simpl.
  extensionality.
  spec H11 x.
  spec H14 x.
  destruct H11.
  destruct H14.
  case (le_gt_dec n x); intro.
  assert (x >= n) by omega.
  spec H H5.
  spec H1 H5.
  congruence.
  assert (x < n) by omega.
  spec H0 H5.
  spec H4 H5.
  congruence.
Qed.

Lemma pred_brk_ubiquitous : ubiquitous pred_brk.
Proof.
  unfold ubiquitous, everywhere.
  extensionality st.
  apply prop_ext; split; intro; auto.
  destruct st as [[? ?] ?].
  spec H s.
  auto.
Qed.

Definition closedWith (F : assert) (c : cmd) : Prop :=
  (forall x,
    modifies c x ->
      forall st v,
        F st -> 
          F (st_upd_stk_var st x v))
(* /\
  (forall bn,
    modifies_bn c bn -> (* not 100% obvious that this is ok... *)
      forall st bst,
        F st ->
          F (st_upd_bm_bn st bn bst))
*)
.

Lemma emp_closedWith: forall c,
  closedWith emp c.
Proof.
  unfold closedWith.
  intros.
  rewrite <- emp_ubiquitous in H0.
  spec H0 (upd_stack (st_stk st) x v).
  apply H0.
Qed.

(* Uggg-ly *)
Lemma joinlistState_joinlistHeap : forall stk bs bs' h (Sx : list state) h',
  join_list (stk, h, bs) Sx (stk, h', bs') ->
  join_list h (map (fun st => snd (fst st)) Sx) h'.
Proof.
  intros until Sx.
  revert h bs bs'.
  induction Sx; intros.
  inversion H; subst.
  constructor.
  inversion H; subst.
  econstructor.
  destruct H3 as [[_ ?] _].
  apply H0.
  destruct w1 as [[? ?] ?].
  destruct a as [[? ?] ?].
  apply IHSx with b bs'.
  destruct H3 as [[[? ?] ?] ?].
  simpl in *.
  subst.
  trivial.
Qed.

Lemma joinlistState_joinlistBm : forall stk bs bs' h (Sx : list state) h',
  join_list (stk, h, bs) Sx (stk, h', bs') ->
  join_list bs (map (fun st => snd st) Sx) bs'.
Proof.
  intros until Sx.
  revert h bs bs'.
  induction Sx; intros.
  inversion H; subst.
  constructor.
  inversion H; subst.
  econstructor.
  destruct H3 as [[_ _] ?].
  apply H0.
  destruct w1 as [[? ?] ?].
  destruct a as [[? ?] ?].
  apply IHSx with h0 h'.
  destruct H3 as [[[? ?] ?] ?].
  simpl in *.
  subst.
  trivial.
Qed.

Lemma joinlistHeap_joinlistBm_joinlistState : forall (stk :stack) (bs : BarstMap) h (Hx : list heap) Bx bs' h',
  join_list h Hx h' ->
  join_list bs Bx bs' ->
  length Hx = length Bx ->
  join_list (stk, h, bs) (map (fun p => (stk, fst p, snd p)) (zip Hx Bx)) (stk, h', bs').
Proof.
  intros until Hx.
  revert h bs.
  induction Hx; intros.
  inversion H; subst.
  icase Bx.
  inversion H0; subst.
  constructor.
  icase Bx.
  inversion H1.
  inversion H; subst.
  inversion H0; subst.
  replace (map (fun p : heap * BarstMap => (stk, fst p, snd p)) (zip (a :: Hx) (b :: Bx))) with
                ((stk, a, b) :: (map (fun p => (stk, fst p, snd p)) (zip Hx Bx))).
  econstructor.
  instantiate (1 := (stk, w1, w0)).
  repeat (split; trivial).
  spec IHHx w1 w0 Bx bs' h'.
  apply IHHx; auto.
  clear -H3.
  unfold zip.
  simpl.
  trivial.
Qed.

Lemma join_list_empty_state : forall L st n st',
  join_list empty_state L st ->
  nth_error L n = Some st' ->
  st_stk st' = empty_stack.
Proof.
  intros.
  unfold empty_state in H.
  remember empty_heap.
  remember empty_bm.
  clear Heqh Heqb.
  revert n st' H0 h b H.
  induction L; intros.
  icase n.
  destruct n.
  simpl in H0.
  inversion H0.
  subst.
  inversion H; subst.
  destruct st' as [[? ?] ?].
  destruct H4 as [[[? ?] _] _].
  auto.
  simpl in H0.
  inversion H; subst.
  destruct w1 as [[? ?] ?].
  destruct H4 as [[[? ?] _] _].
  rewrite H2 in H1.
  simpl in H1.
  subst s.
  spec IHL n st' H0 h0 b0.
  auto.
Qed.

Lemma join_list_stk : forall L st st',
  join_list st L st' ->
  st_stk st = st_stk st'.
Proof.
  induction L; intros.
  inversion H; auto.
  inversion H; subst.
  apply IHL in H5.
  destruct w1 as [[? ?] ?].
  destruct H3 as [[[? ?] _] _].
  destruct st as [[? ?] ?].
  unfold fst in *.
  rewrite H1 in H0.
  rewrite <- H5.
  rewrite H0.
  trivial.
Qed.

Lemma ubiquitous_points2_const : forall e1 e2 sh, ubiquitous (points2 (expConst e1) sh (expConst e2 )).
Proof.
  repeat intro.
  unfold ubiquitous, everywhere.
  extensionality.
  apply prop_ext.
  split.
  repeat intro.
  spec H (st_stk x).
  trivial.
  repeat intro.
  unfold points2,evalExpr in *.
  destruct x as [[? [? ?]] ?].
  unfold st_brk,st_upd_stk,st_hf, st_hp,st_bsm, fst,snd in *.
  trivial.
Qed.

(*
Notation "P  Q" := (and P Q)(at level 200).
Notation "P ||| Q" := (or P Q)(at level 200).
*)

