(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.

(* ORACLE INTERFACE *)

Module Type ORACLES.

Parameter t : Type.

Parameter ora_barDefs : t  -> list BarDef -> Prop.

Parameter step : (t * state * cmd) -> (option (t * state * cmd)) -> Prop.

End ORACLES.

Module SeqStep (O : ORACLES).

Definition config : Type := (state * O.t * cmd)%type.

(* SEQUENTIAL SEMANTICS *)
Inductive seq_step : config -> config -> Prop :=

  | SeqStep_Skip : forall st o k,
(* ------------------------------------------------------------*)
  seq_step (st, o, Seq Unit k) (st, o, k)


  | SeqStep_Seq : forall st o k cmd1 cmd2,
(* ------------------------------------------------------------*)
  seq_step (st, o, Seq (Seq cmd1 cmd2) k) 
                   (st, o, Seq cmd1 (Seq cmd2 k))


  | SeqStep_Assign: forall  stk stk' hp var e o k bs,
  stk' = upd_stack stk var (evalExpr stk e) ->
(* ------------------------------------------------------------*)
  seq_step ((stk, hp, bs), o, Seq (Assign var e) k)
                  ((stk', hp, bs), o, k)


  | SeqStep_Load : forall stk stk' hp var e brk addr val (sh : pshare) o k bs,
  evalExpr stk e = V_addr addr ->
  dref_heap hp addr = res_builder val sh ->
  brk_heap hp = Some brk ->
  addr < brk ->
  stk' = upd_stack stk var val ->
(* ------------------------------------------------------------*)
  seq_step ((stk, hp, bs), o, Seq (Load var e) k)
                   ((stk', hp, bs), o, k)


  | SeqStep_Store : forall stk hp addr brk val val' o k e1 e2 hp' bs,
  evalExpr stk e1 = V_addr addr ->
  evalExpr stk e2 = val ->
  brk_heap hp = Some brk ->
  addr < brk ->
  dref_heap hp addr = res_builder val' pfullshare ->
  hp' = upd_heap hp addr (res_builder val pfullshare) ->
(* ------------------------------------------------------------*)
  seq_step ((stk, hp, bs), o, Seq (Store e1 e2) k) 
                   ((stk, hp', bs), o, k)


  | SeqStep_ifthen : forall stk h e cmd1 cmd2 o k bs,  
  evalExpr stk e = truE->
(* ------------------------------------------------------------*)
  seq_step ((stk, h, bs), o, Seq (If e cmd1 cmd2) k)
                   ((stk, h, bs), o, Seq cmd1 k)


  | SeqStep_ifelse : forall stk h e cmd2 (cmd1:cmd) o k bs,
  ~ evalExpr stk e = truE ->
(* ------------------------------------------------------------*)
  seq_step ((stk, h, bs), o, Seq (If e cmd1 cmd2) k)
                   ((stk, h, bs), o, Seq cmd2 k)


  | SeqStep_WhileTru : forall stk h e cmd' o k bs,
  evalExpr stk e = truE->
(* ------------------------------------------------------------*)
  seq_step ((stk, h, bs), o, Seq (While e cmd') k)
                   ((stk, h, bs), o, Seq cmd' (Seq (While e cmd') k))


  | SeqStep_WhileBrk : forall stk h cmd e o k bs,
  ~(evalExpr stk e = truE) ->
(* ------------------------------------------------------------*)
  seq_step ((stk, h, bs), o, Seq (While e cmd) k) 
                   ((stk, h, bs), o, k)


  | SeqStep_Free : forall (stk : stack) (hp hp' : heap) brk e addr val' o k bs,
  evalExpr stk e = V_addr addr ->
  brk_heap hp = Some brk ->
  addr < brk ->
  dref_heap hp addr = res_builder val' pfullshare ->
  hp' = upd_heap hp addr None ->
(* ------------------------------------------------------------*)
  seq_step ((stk, hp, bs), o, Seq (Free e) k) 
                   ((stk, hp', bs), o, k)


  | SeqStep_New : forall o stk stk' hp hp' v val e k (addr : addr) bs,
  evalExpr stk e = val ->
  new_heap hp (res_builder val pfullshare) = Some (hp', addr) ->
  stk' = upd_stack stk v (V_addr addr) ->
(* ------------------------------------------------------------*)
  seq_step ((stk, hp, bs), o, Seq (New v e)  k)
                   ((stk', hp', bs), o, k)


  | SeqStep_BarStep : forall (st : state) o (b : bar_num) c st' o' c',
  O.step (o, st, Seq (Barrier b ) c) (Some (o', st', c')) ->
(* ------------------------------------------------------------*)
  seq_step (st, o, Seq (Barrier b) c) 
                   (st', o', c')


  | SeqStep_BarLoop : forall st o b c,
  O.step (o, st, Seq (Barrier b) c) None ->
(* ------------------------------------------------------------*)
  seq_step (st, o, Seq (Barrier b) c) 
                   (st, o, Seq (Barrier b) c).

Inductive seq_stepstar : config -> config -> Prop :=
  | SeqStepStar0 : forall c,
     seq_stepstar c c
  | SeqStepStarS : forall c c' c'',
    seq_step c c' ->
    seq_stepstar c' c'' ->
    seq_stepstar c c''.
    
Inductive seq_stepn : nat -> config -> config -> Prop :=
  | SeqStepn0 : forall c,
     seq_stepn 0 c c
  | SeqStepnS : forall c n c' c'',
    seq_step c c' ->
    seq_stepn n c' c'' ->
    seq_stepn (S n) c c''.

Lemma stepstar_stepn : forall c c' ,
	seq_stepstar c c' -> exists n, seq_stepn n c c'.
Proof.
   intros.
   induction H.
   exists 0.
   constructor.
   destruct IHseq_stepstar.
   exists (S x).
   apply SeqStepnS with c'.
   trivial.
   trivial.
Qed.

Lemma stepn_stepstar: forall n c c' ,
	seq_stepn n c c' -> seq_stepstar c c'.
Proof.
  intros.
  induction H.
  constructor.
  apply SeqStepStarS with c'; trivial.
Qed.


Definition safe_halt (c : config) : Prop :=
  match c with (_, _, Unit) => True | _ => False end.

End SeqStep.

Module TrivialOracle <: ORACLES.

Definition t : Type := unit.

Definition ora_barDefs (o : t) (sbs : list BarDef) : Prop :=
  True.

Definition step (x : t * state * cmd) (y : option (t * state * cmd)) : Prop :=
  False.

End TrivialOracle.

Module SeqMachine := SeqStep(TrivialOracle).

Module SM := SeqMachine.

Lemma smsstep_heap: forall s h1 h2 h3 b c s' h3' b' c',
  pred_brk (empty_stack, h2, empty_bm) ->
  join h1 h2 h3 ->
  SM.seq_step ((s,h3,b), tt, c) ((s',h3',b'), tt, c') ->
  exists h1', exists h2',
  sub_silhouette h3' h3 /\
  join h1' h2' h3' /\
  pred_brk (empty_stack, h2', empty_bm).
Proof.
  intros.
  inversion H1; rename H1 into H1cpy; subst; try (exists h1; exists h2; split; try apply sub_silhouette_refl; tauto); contr.
(* store *)
  exists (upd_heap h1 addr (res_builder (evalExpr s' e2) pfullshare)).
  exists h2.
  split.
  do 2 intro.
  destruct H1.
  exists (upd_heap x addr (res_builder (evalExpr s' e2) pfullshare)).
  destruct H1.
  split; trivial.
  intro.
  spec H2 x0.
  unfold upd_heap, fst, snd in *.
  destruct c.
  destruct h3.
  destruct x.
  revert H2.
  case (eq_nat_dec x0 addr); intros; trivial.
  unfold dref_heap, snd in H15.
  subst x0.
  rewrite H15 in H2.
  unfold res_builder in H2.
  simpl in H2.
  inversion H2.
  constructor.
  destruct x.
  destruct z.
  destruct H6.
  destruct x.
  unfold fst, snd in *.
  generalize (nidentity_pair2 n); intro.
  elimtype False.
  destruct x0.
  change t with (projT1 (existT (fun x => ~identity x) _ H8)) in H7.
  change Share.top with (projT1 pfullshare) in H7.
  pfullshare_join.
  split; trivial.
  destruct h1.
  destruct h2.
  destruct h3.
  destruct H0.
  split; unfold upd_heap, fst, snd in *; auto.
  intro.
  spec H1 x.
  unfold dref_heap, snd in H15.
  case (eq_nat_dec x addr); intro; trivial.
  subst x.
  unfold brk_heap, fst in H13.
  subst b1.
  destruct H as [? [? [? ?]]].
  unfold st_brk, st_hf, st_bsm, st_hp, fst, snd in *.
  subst b0.
  icase b.
  compute in H0.
  subst x.
  spec H3 addr.
  destruct H3.
  spec H2 H14.
  rewrite H2.
  constructor.
(* Free *)
  exists (upd_heap h1 addr None).
  exists h2.
  split.
  destruct h3.
  intros [? ?] [[? ?] ?].
  exists (upd_heap (b1, r1) addr None).
  destruct H1; split; unfold brk_heap, dref_heap, upd_heap, fst, snd in *.
  trivial.
  intro adr.
  spec H2 adr.
  case (eq_nat_dec adr addr); intro; trivial.
  subst adr.
  rewrite H14 in H2.
  unfold res_builder in H2.
  inversion H2.
  constructor.
  destruct x.
  destruct z.
  destruct H6.
  destruct x.
  simpl in H7.
  generalize (nidentity_pair2 n); intro.
  elimtype False.
  change t with (projT1 (existT (fun x => ~identity x) _ H9)) in H7.
  change Share.top with (projT1 pfullshare) in H7.
  pfullshare_join.
  split; trivial.
  destruct h1.
  destruct h2.
  destruct h3.
  destruct H0.
  split; unfold upd_heap, fst, snd in *; auto.
  intro.
  unfold dref_heap, snd in H14.
  case (eq_nat_dec x addr); intro; trivial.
  subst x.
  unfold brk_heap, fst in H12.
  subst b1.
  destruct H as [? [? [? ?]]].
  unfold st_brk, st_hf, st_bsm, st_hp, fst, snd in *.
  subst b0.
  icase b.
  compute in H0.
  subst x.
  spec H3 addr.
  destruct H3.
  spec H2 H13.
  rewrite H2.
  constructor.
(* new *)
  destruct h1.
  destruct h2.
  destruct h3.
  destruct H0.
  unfold new_heap, fst, snd in *.
  icase b1.
  inversion H12; subst.
  clear H12.
  destruct H as [? [? [? ?]]].
  unfold st_brk, st_hf, st_bsm, st_hp, fst, snd in *.
  subst b0.
  icase b.
  compute in H0.
  subst x.
  exists (upd_heap (None, r) addr (res_builder (evalExpr s e) pfullshare)).
  exists (upd_heap (Some (addr + 1), r0) addr None).
  split.
  intros [? ?] [[? ?] [? ?]].
  exists (upd_heap (Some (addr + 1), r3) addr (res_builder (evalExpr s e) pfullshare)).
  spec H3 addr.
  destruct H3.
  assert (addr >= addr) by omega.
  spec H3 H5.
  split.
  simpl.
  icase b.
  reflexivity.
  intro adr.
  spec H2 adr.
  unfold upd_heap, snd in *.
  case (eq_nat_dec adr addr); intro; trivial.
  subst adr.
  spec H1 addr.
  rewrite H3 in H1.
  unfold res_builder in H1.
  inversion H1; subst.
  rewrite <- H9 in H2.
  inversion H2; subst.
  constructor.
  destruct x.
  destruct z.
  destruct H11.
  destruct x.
  destruct x0.
  unfold snd in *.
  generalize (nidentity_pair2 n); intro.
  elimtype False.
  change t with (projT1 (existT (fun x => ~identity x) _ H12)) in H11.
  change Share.top with (projT1 pfullshare) in H11.
  pfullshare_join.
  destruct x.
  destruct z.
  destruct H9.
  destruct x.
  destruct x0.
  unfold snd in *.
  generalize (nidentity_pair2 n); intro.
  elimtype False.
  change t with (projT1 (existT (fun x => ~identity x) _ H10)) in H9.
  change Share.top with (projT1 pfullshare) in H9.
  pfullshare_join.
  split.
  split.
  constructor.
  intro.
  unfold upd_heap, fst, snd.
  case (eq_nat_dec x addr); intro.
  subst x.
  constructor.
  spec H1 x.
  trivial.
  split.
  trivial.
  exists (addr + 1).
  split; trivial.
  intro.
  unfold st_hf, st_hp, fst, snd, upd_heap.
  split; intro.
  case (eq_nat_dec n addr); intro.
  subst n.
  omegac.
  spec H3 n.
  destruct H3.
  simpl.
  rewrite H2.
  trivial.
  omega.
  case (eq_nat_dec n addr); intro.
  trivial.
  assert (n < addr) by omega.
  spec H3 n.
  destruct H3.
  spec H4 H2.
  simpl.
  trivial.
Qed.

Lemma smsstep_bs: forall s h b c s' h' b' c',
  SM.seq_step ((s,h,b), tt, c) ((s', h', b'), tt, c') ->
  b = b'.
Proof.
  inversion 1; contr; subst; auto.
Qed.
